"""Implements :class:`DataArray`-like objects for handling arrayed data.

:class:`DataArray`-like objects are concrete implementations of :class:`.ArrayBase` base
class that collect specific data for multiple conformers and provide an easy access to
genre-specific functionality. Instances of :class:`DataArray` subclasses are produced by
the :meth:`.Conformers.arrayed` method and :class:`Tesliper`'s subscription mechanism.
"""

import logging as lgg
from abc import ABC, abstractmethod

# IMPORTS
from inspect import Parameter
from typing import Any, Dict, Sequence, Tuple, Union

import numpy as np

from .. import datawork as dw
from ..datawork import convert_band
from ..datawork.atoms import atomic_number
from .array_base import (
    ArrayBase,
    ArrayProperty,
    CollapsibleArrayProperty,
    DependentParameter,
    JaggedArrayProperty,
)
from .spectra import Spectra

# LOGGER
logger = lgg.getLogger(__name__)
logger.setLevel(lgg.DEBUG)


# CLASSES
class DataArray(ArrayBase):
    """Base class for data holding objects."""

    associated_genres = tuple()
    _full_name_ref = {}
    _units = {}

    @property
    def full_name(self):
        try:
            return self._full_name_ref[self.genre]
        except KeyError:
            return f"{self.genre} data"

    @property
    def units(self):
        try:
            return self._units[self.genre]
        except KeyError:
            return ""


class IntegerArray(DataArray):
    """For handling data of ``int`` type.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - charge
          - multiplicity
    """

    associated_genres = ("charge", "multiplicity")
    _full_name_ref = {"charge": "Charge", "multiplicity": "Multiplicity"}
    values = ArrayProperty(dtype=int, check_against="filenames")


class FloatArray(DataArray):
    """For handling data of ``float`` type.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - zpecorr
          - tencorr
          - entcorr
          - gibcorr

    """

    associated_genres = (
        "zpecorr",
        "tencorr",
        "entcorr",
        "gibcorr",
    )
    _full_name_ref = dict(
        zpecorr="Zero-point Correction",
        tencorr="Correction to Energy",
        entcorr="Correction to Enthalpy",
        gibcorr="Correction to Free Energy",
    )
    _units = dict(
        zpecorr="Hartree",
        tencorr="Hartree",
        entcorr="Hartree",
        gibcorr="Hartree",
    )

    values = ArrayProperty(dtype=float, check_against="filenames")


class InfoArray(DataArray):
    """For handling data of ``str`` type.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - command
          - stoichiometry
    """

    _full_name_ref = {
        "command": "Command",
        "stoichiometry": "Stoichiometry",
    }
    _units = {}
    associated_genres = (
        "command",
        "stoichiometry",
    )
    values = ArrayProperty(dtype=str, check_against="filenames")


class FilenamesArray(DataArray):
    """Special case of :class:`DataArray`, holds only filenames. *values* property
    returns same as *filenames* and ignores any value given to its setter.
    Only genre associated with this class is *filenames* pseudo-genre.
    """

    associated_genres = ("filenames",)
    _full_name_ref = {"filenames": "Filenames"}
    _units = {}

    def __init__(
        self,
        genre: str = "filenames",
        filenames: Union[Sequence, np.ndarray] = (),
        values: Any = None,
        allow_data_inconsistency: bool = False,
    ):
        """
        Parameters
        ----------
        genre : str
            Name of genre, should be 'filenames'.
        filenames : numpy.ndarray(dtype=str)
            List of filenames of gaussian output files, from which data were extracted.
        values : numpy.ndarray(dtype=str)
            Always returns same as *filenames*.
        """
        super().__init__(genre, filenames, values, allow_data_inconsistency)

    @property
    def values(self):
        return self.filenames

    @values.setter
    def values(self, values):
        # ignore attempts at setting values
        pass


class BooleanArray(DataArray):
    """For handling data of ``bool`` type.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - normal_termination
          - optimization_completed
    """

    _full_name_ref = {
        "normal_termination": "Normal Termination",
        "optimization_completed": "Optimization Completed",
    }
    _units = {}
    associated_genres = ("normal_termination", "optimization_completed")
    values = ArrayProperty(dtype=bool, check_against="filenames")


class Energies(FloatArray):
    """For handling data about the energy of conformers.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - scf
          - zpe
          - ten
          - ent
          - gib
    """

    _full_name_ref = dict(
        zpe="Zero-point Energy",
        ten="Thermal Energy",
        ent="Thermal Enthalpy",
        gib="Thermal Free Energy",
        scf="SCF",
    )
    _units = dict(
        zpe="Hartree",
        ten="Hartree",
        ent="Hartree",
        gib="Hartree",
        scf="Hartree",
    )
    associated_genres = (
        "scf",
        "zpe",
        "ten",
        "ent",
        "gib",
    )

    def __init__(
        self, genre, filenames, values, t=298.15, allow_data_inconsistency=False
    ):
        """
        Parameters
        ----------
        genre : str
            genre of energy.
        filenames : numpy.ndarray(dtype=str)
            List of filenames of gaussian output files, from which data were
            extracted.
        values : numpy.ndarray(dtype=float)
            Energy value for each conformer.
        t : int or float
            Temperature of calculated state in K.
        """
        super().__init__(genre, filenames, values, allow_data_inconsistency)
        self.t = t  # temperature in K

    @property
    def as_kcal_per_mol(self):
        """Energy values converted to kcal/mol."""
        # convert hartree to kcal/mol by multiplying by 627.5095
        return self.values * dw.energies.HARTREE_TO_KCAL_PER_MOL

    @property
    def deltas(self):
        """Calculates energy difference between each conformer and lowest energy
        conformer. Converts energy to kcal/mol.

        Returns
        -------
        numpy.ndarray
            List of energy differences from lowest energy in kcal/mol."""
        return dw.calculate_deltas(self.as_kcal_per_mol)

    @property
    def min_factors(self):
        """Calculates list of conformers' Boltzmann factors respective to lowest
        energy conformer in system.

        Notes
        -----
        Boltzmann factor of two states is defined as:
        F(state_1)/F(state_2) = exp((E_1 - E_2)/kt)
        where E_1 and E_2 are energies of states 1 and 2,
        k is Boltzmann constant, k = 0.0019872041 kcal/(mol*K),
        and t is temperature of the system.

        Returns
        -------
        numpy.ndarary
            List of conformers' Boltzmann factors respective to lowest
            energy conformer."""
        # F(state_n)/F(state_min)
        return dw.calculate_min_factors(self.as_kcal_per_mol, self.t)

    @property
    def populations(self):
        """Calculates Boltzmann distribution of conformers.

        Returns
        -------
        numpy.ndarary
            List of conformers populations calculated as Boltzmann
            distribution."""
        return dw.calculate_populations(self.as_kcal_per_mol, self.t)

    def calculate_populations(self, t):
        """Calculates conformers' Boltzmann distribution in given temperature.

        Parameters
        ----------
        t : int or float
            Temperature of calculated state in K."""
        return dw.calculate_populations(self.as_kcal_per_mol, t)


class Averagable:
    """Mix-in for DataArray subclasses, that may be averaged based on populations
    of conformers."""

    def average_conformers(self: DataArray, energies) -> DataArray:
        """A method for averaging values by population of conformers.

        Parameters
        ----------
        energies : Energies or iterable
            Object with ``populations`` and ``genre`` attributes, containing
            respectively: list of populations values as numpy.ndarray and
            string specifying energy type. Alternatively, list of weights
            for each conformer.

        Returns
        -------
        DataArray
            New instance of DataArray's subclass, on which *average* method was
            called, containing averaged values.

        Raises
        ------
        TypeError
            If creation of an instance based on its __init__ signature is impossible.
        """
        # TODO: make sure returning DataArray is necessary and beneficial
        #       maybe it should return just averaged value
        try:
            populations = energies.populations
            energy_type = energies.genre
        except AttributeError:
            populations = np.asanyarray(energies, dtype=float)
            energy_type = "unknown"
        averaged_values = dw.calculate_average(self.values, populations)
        args = self.get_repr_args()
        args["values"] = [averaged_values]
        args["allow_data_inconsistency"] = True
        try:
            averaged = type(self)(**args)
        except (TypeError, ValueError) as err:
            raise TypeError(
                f"Could not create an instance of {type(self)} from its "
                f"signature. Use tesliper.datawork.calculate_average instead."
            ) from err
        logger.debug(f"{self.genre} averaged by {energy_type}.")
        return averaged


class Bands(FloatArray):
    """Special kind of data array for band values, to which spectral data or activities
    corespond. Provides an easy way to convert values between their different
    representations: frequency, wavelength, and excitation energy.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - freq
          - wavelen
          - ex_en
    """

    associated_genres = ("freq", "wavelen", "ex_en")
    _full_name_ref = {
        "ex_en": "Excitation energy",
        "freq": "Frequency",
        "wavelen": "Wavelength",
    }
    _units = {"freq": "cm^(-1)", "wavelen": "nm", "ex_en": "eV"}

    @property
    def freq(self):
        """Values converted to frequencies in :math:`\\mathrm{cm}^{-1}`."""
        return convert_band(self.values, from_genre=self.genre, to_genre="freq")

    @property
    def frequencies(self):
        """Values converted to frequencies in :math:`\\mathrm{cm}^{-1}`.
        A convenience alias for :attr:`Bands.frequencies`.
        """
        return self.freq

    @property
    def wavelen(self):
        """Values converted to wavelengths in nm."""
        return convert_band(self.values, from_genre=self.genre, to_genre="wavelen")

    @property
    def wavelengths(self):
        """Values converted to wavelengths in nm.
        A convenience alias for :attr:`Bands.wavelen`.
        """
        return self.wavelen

    @property
    def ex_en(self):
        """Values converted to excitation energy in eV."""
        return convert_band(self.values, from_genre=self.genre, to_genre="ex_en")

    @property
    def excitation_energy(self):
        """Values converted to excitation energy in eV.
        A convenience alias for :attr:`Bands.ex_en`.
        """
        return self.ex_en

    @property
    def imaginary(self):
        """Finds number of imaginary frequencies of each conformer.

        Returns
        -------
        numpy.ndarray
            Number of imaginary frequencies of each conformer."""
        if self.frequencies.size > 0:
            return (self.frequencies < 0).sum(1)
        else:
            return np.array([])

    def find_imaginary(self):
        """Reports number of imaginary frequencies of each conformer that has any.

        Returns
        -------
        dict
            Dictionary of {filename: number-of-imaginary-frequencies} for each
            conformer with at least one imaginary frequency.
        """
        imag = self.imaginary
        return {k: v for k, v in zip(self.filenames, imag) if v}


class SpectralData(FloatArray, ABC):
    """Base class for spectral data genres, that are not spectral activities.

    When subclassed, one of the attributes: :attr:`~.SpectralData.freq` or
    :attr:`~.SpectralData.wavelen` should be overridden with a concrete setter and
    getter - use of :class:`.ArrayProperty` is recommended. The other one may use
    implementation from this base class by call to ``super().freq`` or
    ``super().wavelen`` to get converted values.
    """

    # TODO: Supplement tests regarding this class' subclasses

    associated_genres = ()

    @property
    @abstractmethod
    def spectra_type(self):
        """Type of spectra, that genres associated with :class:`.SpectralData`'s
        subclass relate to. Should be a class-level attribute with value of either
        "vibrational", "electronic", or "scattering".
        """
        return NotImplemented

    @property
    @abstractmethod
    def freq(self):
        """Bands values converted to frequencies in :math:`\\mathrm{cm}^{-1}`. If
        :attr:`~.SpectralData.wavelen` is provided, this may be overridden with a simple
        call to ``super()``:

        .. code-block:: python

            @property
            def freq(self):
                return super().freq()  # values converted to cm^(-1)
        """
        return convert_band(self.wavelen, from_genre="wavelen", to_genre="freq")

    @property
    def frequencies(self):
        """Bands values converted to frequencies in :math:`\\mathrm{cm}^{-1}`.
        A convenience alias for :attr:`~.SpectralData.freq`.
        """
        return self.freq

    @property
    @abstractmethod
    def wavelen(self):
        """Bands values converted to wavelengths in nm. If :attr:`.freq` is
        provided, this may be overridden with a simple call to ``super()``:

        .. code-block:: python

            @property
            def wavelen(self):
                return super().wavelen()  # values converted to nm
        """
        return convert_band(self.freq, from_genre="freq", to_genre="wavelen")

    @property
    def wavelengths(self):
        """Bands values converted to wavelengths in nm.
        A convenience alias for :attr:`~.SpectralData.wavelen`.
        """
        return self.wavelen


class _VibData(SpectralData):
    freq = ArrayProperty(check_against="filenames")
    associated_genres = ()

    def __init__(
        self,
        genre,
        filenames,
        values,
        freq,
        allow_data_inconsistency=False,
    ):
        """
        Parameters
        ----------
        genre
            Name of the data genre that *values* represent.
        filenames
            Sequence of conformers' identifiers.
        values
            Sequence of values for *genre* for each conformer in *filenames*.
        freq
            Frequency for each value in each conformer in :math:`\\mathrm{cm}^{-1}`
            units.
        allow_data_inconsistency
            Flag signalizing if instance should allow data inconsistency (see
            :class:`ArrayPropety` for details).
        """
        super().__init__(genre, filenames, values, allow_data_inconsistency)
        self.freq = freq

    @property
    def wavelen(self):
        """Bands values converted to wavelengths in nm."""
        return super().wavelen


class VibrationalData(_VibData):
    """For handling vibrational data that is not a spectral activity.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - mass
          - frc
          - emang
    """

    associated_genres = ("mass", "frc", "emang")
    _full_name_ref = dict(
        mass="Reduced masses", frc="Force constants", emang="E-M Angle"
    )
    _units = dict(mass="AMU", frc="mDyne/A", emang="deg")

    @property
    def spectra_type(self):
        return "vibrational"


class ScatteringData(_VibData):
    """For handling scattering data that is not a spectral activity.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - depolarp
          - depolaru
          - depp
          - depu
          - alpha2
        * - beta2
          - alphag
          - gamma2
          - delta2
          - cid1
        * - cid2
          - cid3
          - rc180
          -
          -
    """

    associated_genres = (
        "depolarp",
        "depolaru",
        "depp",
        "depu",
        "alpha2",
        "beta2",
        "alphag",
        "gamma2",
        "delta2",
        "cid1",
        "cid2",
        "cid3",
        "rc180",
    )
    _full_name_ref = {
        "depolarp": "Depolar-P Raman",
        "depolaru": "Depolar-U Raman",
        "depp": "Depolar-P ROA",
        "depu": "Depolar-U ROA",
        "alpha2": "Raman invariant Alpha2",
        "beta2": "Raman invariant Beta2",
        "alphag": "ROA invariant AlphaG",
        "gamma2": "ROA invariant Gamma2",
        "delta2": "ROA invariant Delta2",
        "cid1": "CID ICPu/SCPu(180)",
        "cid2": "CID ICPd/SCPd(90)",
        "cid3": "CID DCPI(180)",
        "rc180": "Degree of circularity",
    }
    _units = {
        "alpha2": "(A**4/AMU)",
        "beta2": "(A**4/AMU)",
        "alphag": "(10**4 A**5/AMU)",
        "gamma2": "(10**4 A**5/AMU)",
        "delta2": "(10**4 A**5/AMU)",
    }

    @property
    def spectra_type(self):
        return "scattering"

    def __init__(
        self,
        genre,
        filenames,
        values,
        freq,
        t=298.15,
        laser=532,
        allow_data_inconsistency=False,
    ):
        super().__init__(genre, filenames, values, freq, allow_data_inconsistency)
        self.laser = laser  # in nm
        self.t = t  # temperature in K


class ElectronicData(SpectralData):
    """For handling electronic data that is not a spectral activity.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - eemang
    """

    wavelen = ArrayProperty(check_against="filenames")
    associated_genres = ("eemang",)
    _full_name_ref = dict(eemang="E-M Angle")
    _units = dict(eemang="deg")

    @property
    def spectra_type(self):
        return "electronic"

    def __init__(
        self,
        genre,
        filenames,
        values,
        wavelen,
        allow_data_inconsistency=False,
    ):
        super().__init__(genre, filenames, values, allow_data_inconsistency)
        self.wavelen = wavelen  # in nm

    @property
    def freq(self):
        return super().freq


class SpectralActivities(SpectralData, Averagable, ABC):
    """Base class for spectral activities genres."""

    associated_genres = ()
    spectra_name_ref = dict(
        rot="vcd",
        dip="ir",
        iri="ir",
        ramact="raman",
        ramanactiv="raman",
        raman1="raman",
        roa1="roa",
        raman2="raman",
        roa2="roa",
        raman3="raman",
        roa3="roa",
        vrot="ecd",
        lrot="ecd",
        vosc="uv",
        losc="uv",
        vdip="uv",
        ldip="uv",
    )
    _full_name_ref = dict()
    _units = dict()
    _intensities_converters = {}

    @property
    def spectra_name(self):
        if self.genre in self.spectra_name_ref:
            return self.spectra_name_ref[self.genre]

    @property
    def intensities(self):
        """Converts spectral activity calculated by quantum chemistry software
        to signal intensity.

        Returns
        -------
        numpy.ndarray
            Signal intensities for each conformer.

        Raises
        ------
        NotImplementedError
            if genre does not provide values conversion to intensities."""
        try:
            converter = self._intensities_converters[self.genre]
        except KeyError:
            raise NotImplementedError(
                f"Genre {self.genre} does not provide conversion to intensities."
            )
        return converter(self.values, self.frequencies)

    @abstractmethod
    def calculate_spectra(self, start, stop, step, width, fitting):
        return NotImplemented


def _as_is(values, *_args, **_kwargs):
    return values


class _VibAct(_VibData, SpectralActivities):
    def calculate_spectra(self, start, stop, step, width, fitting):
        """Calculates spectrum for each individual conformer.

        Parameters
        ----------
        start : int or float
            Number representing start of spectral range in relevant units.
        stop : int or float
            Number representing end of spectral range in relevant units.
        step : int or float
            Number representing step of spectral range in relevant units.
        width : int or float
            Number representing half width of maximum peak height.
        fitting : function
            Function, which takes spectral data, freqs, abscissa, width as parameters
            and returns numpy.array of calculated, non-corrected spectrum points.

        Returns
        -------
        SingleSpectrum
            Calculated spectrum.

        Raises
        ------
        ValueError
            If given *start*, *stop*, and *step* values would produce an empty
            or one-element sequence; i.e. if *start* is grater than *stop* or if
            ``start - stop < step``, assuming *step* is a positive value.
        """
        abscissa = np.arange(start, stop, step)
        if abscissa.size <= 1:
            raise ValueError(
                f"Not enough data points between start = {start} and stop = {stop} "
                f"with step = {step}. Please provide start, stop, and step values "
                "that will produce at least two-element sequence."
            )
        freqs = self.frequencies
        inten = self.intensities
        values = dw.calculate_spectra(freqs, inten, abscissa, width, fitting)
        spectra_name = self.spectra_name
        fitting_name = fitting.__name__
        if values.size:
            logger.debug(
                f"Bar {self.genre}: {spectra_name} spectra calculated with "
                f"width = {width} and {fitting_name} fitting."
            )
        spectra = Spectra(
            spectra_name, self.filenames, values, abscissa, width, fitting_name
        )
        return spectra


class VibrationalActivities(VibrationalData, _VibAct):
    """For handling electronic spectral activity data.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - iri
          - dip
          - rot
    """

    associated_genres = (
        "iri",
        "dip",
        "rot",
    )
    _intensities_converters = {
        "dip": dw.dip_to_ir,
        "rot": dw.rot_to_vcd,
        "iri": _as_is,
    }
    _full_name_ref = dict(
        rot="Rot. Strength",
        dip="Dip. Strength",
        iri="IR Intensity",
    )
    _units = dict(
        rot="10^(-44) esu^2 cm^2",
        dip="10^(-40) esu^2 cm^2",
        iri="KM/Mole",
    )


class ScatteringActivities(ScatteringData, _VibAct):
    """For handling scattering spectral activity data.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - ramanactiv
          - ramact
          - raman1
          - roa1
        * - raman2
          - roa2
          - raman3
          - roa3
    """

    associated_genres = (
        "ramanactiv",
        "ramact",
        "raman1",
        "roa1",
        "raman2",
        "roa2",
        "raman3",
        "roa3",
    )
    _full_name_ref = dict(
        ramanactiv="Raman scatt. activities",
        ramact="Raman scatt. activities",
        roa1="ROA inten. ICPu/SCPu(180)",
        raman1="Raman inten. ICPu/SCPu(180)",
        roa2="ROA inten. ICPd/SCPd(90)",
        raman2="Raman inten. ICPd/SCPd(90)",
        roa3="ROA inten. DCPI(180)",
        raman3="Raman inten. DCPI(180)",
    )
    _units = dict(
        ramanactiv="A^4/AMU",
        ramact="A^4/AMU",
        roa1="10^4 K",
        raman1="K",
        roa2="10^4 K",
        raman2="K",
        roa3="10^4 K",
        raman3="K",
    )
    _intensities_converters = {
        "ramanactiv": _as_is,
        "ramact": _as_is,
        "raman1": _as_is,
        "roa1": _as_is,
        "raman2": _as_is,
        "roa2": _as_is,
        "raman3": _as_is,
        "roa3": _as_is,
    }

    @property
    def intensities(self):
        """Converts spectral activity calculated by quantum chemistry software
        to signal intensity.

        Returns
        -------
        numpy.ndarray
            Signal intensities for each conformer.

        Raises
        ------
        NotImplementedError
            if genre does not provide values conversion to intensities."""
        try:
            converter = self._intensities_converters[self.genre]
        except KeyError:
            return super().intensities
        return converter(self.values, self.frequencies, self.t, self.laser)


class ElectronicActivities(ElectronicData, SpectralActivities):
    """For handling electronic spectral activity data.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - vdip
          - ldip
          - vrot
          - lrot
          - vosc
          - losc
    """

    associated_genres = (
        "vdip",
        "ldip",
        "vrot",
        "lrot",
        "vosc",
        "losc",
    )
    _full_name_ref = dict(
        vrot="Rot. (velo)",
        lrot="Rot. (lenght)",
        vosc="Osc. (velo)",
        losc="Osc. (length)",
        vdip="Dip. (velo)",
        ldip="Dip. (length)",
    )
    _units = dict(
        vrot="10^(-40) erg*esu*cm/Gauss",
        lrot="10^(-40) erg*esu*cm/Gauss",
        vdip="10^(-44) esu^2 cm^2",
        ldip="10^(-44) esu^2 cm^2",
    )
    _intensities_converters = {
        # for "osc" ignore frequencies given by default by self.intensities
        "vosc": lambda v, _: dw.osc_to_uv(v),
        "losc": lambda v, _: dw.osc_to_uv(v),
        "vrot": dw.rot_to_ecd,
        "lrot": dw.rot_to_ecd,
        "ldip": dw.dip_to_uv,
        "vdip": dw.dip_to_uv,
    }

    @property
    def intensities(self):
        """Converts spectral activity calculated by quantum chemistry software
        to signal intensity.

        Returns
        -------
        numpy.ndarray
            Signal intensities for each conformer.

        Raises
        ------
        NotImplementedError
            if genre does not provide values conversion to intensities."""
        try:
            converter = self._intensities_converters[self.genre]
        except KeyError:
            raise NotImplementedError(
                f"Genre {self.genre} does not provide conversion to intensities."
            )
        return converter(self.values, self.wavelengths)

    def calculate_spectra(self, start, stop, step, width, fitting):
        """Calculates spectrum for each individual conformer.

        Parameters
        ----------
        start : int or float
            Number representing start of spectral range in relevant units.
        stop : int or float
            Number representing end of spectral range in relevant units.
        step : int or float
            Number representing step of spectral range in relevant units.
        width : int or float
            Number representing half width of maximum peak height.
        fitting : function
            Function, which takes spectral data, freqs, abscissa, width as parameters
            and returns numpy.array of calculated, non-corrected spectrum points.

        Returns
        -------
        SingleSpectrum
            Calculated spectrum.

        Raises
        ------
        ValueError
            If given *start*, *stop*, and *step* values would produce an empty
            or one-element sequence; i.e. if *start* is grater than *stop* or if
            ``start - stop < step``, assuming *step* is a positive value.
        """
        abscissa = np.arange(start, stop, step)
        if abscissa.size <= 1:
            raise ValueError(
                f"Not enough data points between start = {start} and stop = {stop} "
                f"with step = {step}. Please provide start, stop, and step values "
                "that will produce at least two-element sequence."
            )
        _width = convert_band(
            width, from_genre="ex_en", to_genre="freq"
        )  # from eV to cm-1
        _abscissa = convert_band(
            abscissa, from_genre="wavelen", to_genre="freq"
        )  # from nm to cm-1
        freqs = self.frequencies
        inten = self.intensities
        values = dw.calculate_spectra(freqs, inten, _abscissa, _width, fitting)
        spectra_name = self.spectra_name
        fitting_name = fitting.__name__
        if values.size:
            logger.debug(
                f"Bar {self.genre}: {spectra_name} spectra calculated with "
                f"width = {width} and {fitting_name} fitting."
            )
        spectra = Spectra(
            spectra_name, self.filenames, values, abscissa, width, fitting_name
        )
        return spectra


class Transitions(DataArray):
    """For handling information about electronic transitions from ground
    to excited state contributing to each band.

    Data is stored in three attributes: :attr:`.ground`, :attr:`.excited`, and
    :attr:`.values`, which are respectively: list of ground state electronic subshells,
    list of excited state electronic subshells, and list of coefficients of transitions
    from corresponding ground to excited subshell. Each of these arrays is of shape
    (conformers, bands, max_transitions), where 'max_transitions' is a highest number of
    transitions contributing to single band across all bands of all conformers.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - transitions

    Attributes
    ----------
    values : numpy.ndarray(dtype=float)
        List of coefficients of each transition. It is a 3-dimensional of shape
        (conformers, bands, max_transitions).
    ground : numpy.ndarray(dtype=int)
        List of ground state electronic subshells, stored as integers assigned to them
        by used quantum computations program. It is a 3-dimensional array of shape
        (conformers, bands, max_transitions).
    excited : numpy.ndarray(dtype=int)
        List of excited state electronic subshells, stored as integers assigned to them
        by used quantum computations program. It is a 3-dimensional array of shape
        (conformers, bands, max_transitions).
    """

    associated_genres = ("transitions",)
    _full_name_ref = dict(transitions="Transitions")
    _units = dict()
    ground = JaggedArrayProperty(dtype=int, check_against="filenames")
    excited = JaggedArrayProperty(dtype=int, check_against="filenames")
    values = JaggedArrayProperty(dtype=float, check_against="filenames")

    @staticmethod
    def unpack_values(values: Sequence[Sequence[Sequence[Tuple[int, int, float]]]]):
        """Unpack transitions data stored as list of tuples of (ground, excited,
        coefficient) to separate lists for each information pice, keeping original
        dimensionality (conformers, bands, transitions).

        Parameters
        ----------
        values : list of lists of lists of tuples of (int, int, float)
            Transitions data (ground and excited state electronic subshell and
            coefficient of transition from former to latter) for each transition
            of each band of each conformer.

        Returns
        -------
        list of lists of lists of int,
        list of lists of lists of int,
        list of lists of lists of float
            Transitions data separated to lists of ground, excited, and coefficients,
            for each transition of each band of each conformer.
        """
        outs = [[], [], []]  # ground, excited, coefs
        for conformer in values:
            [out.append(list()) for out in outs]
            curr_confs = [out[-1] for out in outs]
            for band in conformer:
                [c.append(list()) for c in curr_confs]
                curr_bands = [c[-1] for c in curr_confs]
                for transition in band:
                    for container, value in zip(curr_bands, transition):
                        container.append(value)
        return outs

    def __init__(
        self,
        genre: str,
        filenames: Sequence[str],
        values: Sequence[Sequence[Sequence[Tuple[int, int, float]]]],
        allow_data_inconsistency: bool = False,
    ):
        """
        Parameters
        ----------
        genre
            Name of the data genre that *values* represent.
        filenames
            Sequence of conformers' identifiers.
        values : list of lists of lists of tuples of (int, int, float)
            Transitions data (ground and excited state electronic subshell and
            coefficient of transition from former to latter) for each transition
            of each band of each conformer.
        allow_data_inconsistency
            Flag signalizing if instance should allow data inconsistency (see
            :class:`ArrayPropety` for details).
        """
        super().__init__(genre, filenames, values, allow_data_inconsistency)
        ground, excited, values = self.unpack_values(values)
        self.ground = ground
        self.excited = excited
        self.values = values

    @property
    def coefficients(self) -> np.ndarray:
        """Coefficients of each transition, alias for *values*."""
        return self.values

    @coefficients.setter
    def coefficients(self, values):
        self.values = values

    @property
    def contribution(self) -> np.ndarray:
        """Contribution of each transition to given band, calculated as 2 * coef^2.
        To get values in percent, multiply by 100."""
        return 2 * np.square(self.values)

    @property
    def indices_highest(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Indices of coefficients of highest contribution to band in form that
        can be used in numpy's advanced indexing mechanism."""
        contribution = self.contribution
        indices = contribution.argmax(axis=2)
        x, y, _ = contribution.shape
        # np.ogrid generates missing part of a slice tuple; i.e. creates
        # arrays of integers from 0 to n, with appropriate dimensionality,
        # where n is size of given dimension
        x, y = np.ogrid[:x, :y]
        # returned tuple can be used to slice original values array
        return x, y, indices

    @property
    def highest_contribution(
        self,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Electronic transitions data limited to transition of highest contribution
        to each band. Returns tuple with 4 arrays: ground and excited state electronic
        subshell, coefficient of transition from former to latter, and its contribution,
        for each band of each conformer."""
        indices = self.indices_highest
        # could be also achieved by the following:
        # np.take_along_axis(values, indices[..., np.newaxis], axis=2).squeeze(axis=2)
        # but indexing is much quicker, once *indices* is established
        return (
            self.ground[indices],
            self.excited[indices],
            self.values[indices],
            self.contribution[indices],
        )


def _geom_to_atoms_genre(genre):
    return genre.replace("geom", "atoms")


class Geometry(FloatArray):
    """For handling information about geometry of conformers.

    .. list-table:: Genres associated with this class:
        :width: 100%

        * - last_read_geom
          - input_geom
          - optimized_geom
    """

    associated_genres = ("last_read_geom", "input_geom", "optimized_geom")
    _full_name_ref = dict(
        last_read_geom="Geometry",
        input_geom="Input Geometry",
        optimized_geom="Optimized Geometry",
    )
    _units = dict(
        last_read_geom="Angstrom", input_geom="Angstrom", optimized_geom="Angstrom"
    )
    values = ArrayProperty(dtype=float, check_against="filenames")
    atoms = CollapsibleArrayProperty(
        dtype=int,
        check_against="values",
        check_depth=2,
        # TODO: make sanitizer, that accepts jagged nested sequences
        fsan=np.vectorize(atomic_number, otypes=[int]),
        strict=True,
    )

    @classmethod
    def get_init_params(cls) -> Dict[str, Union[str, Parameter, DependentParameter]]:
        params = super().get_init_params()
        params["atoms"] = DependentParameter.from_parameter(
            params["atoms"], genre_getter=_geom_to_atoms_genre
        )
        return params

    def __init__(
        self,
        genre: str,
        filenames: Sequence[str],
        values: Sequence[Sequence[Sequence[float]]],
        atoms: Union[Sequence[Union[int, str]], Sequence[Sequence[Union[int, str]]]],
        allow_data_inconsistency: bool = False,
    ):
        """
        Parameters
        ----------
        genre
            Name of the data genre that *values* represent.
        filenames
            Sequence of conformers' identifiers.
        values
            List of x, y, z coordinated for each conformer, for each atom.
        allow_data_inconsistency
            Flag signalizing if instance should allow data inconsistency (see
            :class:`ArrayPropety` for details). False by default.
        atoms
            List of atomic numbers representing atoms in conformer, one for each
            coordinate. Should be a list of integers or list of strings, that can be
            interpreted as integers or symbols of atoms. May also be a list of such
            lists - one list of atoms for each conformer. All those lists should be
            identical in such case, otherwise InconsistentDataError is raised. Only one
            list of atoms is stored in either case.
        """
        super().__init__(genre, filenames, values, allow_data_inconsistency)
        self.atoms = atoms



"""Project-specific errors."""


class TesliperError(Exception):
    """Base class for Exceptions used by tesliper library."""

    pass


class InconsistentDataError(TesliperError):
    """Raised to signalize problems with conformers' data consistency.
    Subclasses TesliperError."""

    pass


class InvalidStateError(TesliperError, ValueError):
    """Used by ParserBase class to signalize problems when handling states.
    Subclasses TesliperError and ValueError."""

    pass


class InvalidElementError(TesliperError, ValueError):
    """Used by tesliper to indicate, that value cannot be interpreted as an
    element. Subclasses TesliperError and ValueError."""

    pass



from unittest import mock

import numpy as np
import pytest

import tesliper.glassware.arrays as ar
from tesliper.exceptions import InconsistentDataError
from tesliper.glassware.array_base import _ARRAY_CONSTRUCTORS


def test_filenames_array_empty():
    arr = ar.FilenamesArray()
    assert arr.filenames.tolist() == []
    assert arr.genre == "filenames"


def test_filenames_array_filenames():
    arr = ar.FilenamesArray(filenames=["one", "two"])
    assert arr.filenames.tolist() == ["one", "two"]


def test_filenames_array_values():
    arr = ar.FilenamesArray(filenames=["one", "two"])
    assert arr.filenames is arr.values


def test_dtype():
    arr = ar.FloatArray(genre="bla", filenames=["f1", "f2", "f3"], values=[3, 12, 15])
    assert arr.values.dtype == np.float64


@pytest.mark.parametrize(
    "cls,kwargs",
    [
        (ar.IntegerArray, {}),
        (ar.FloatArray, {}),
        (ar.InfoArray, {}),
        (ar.FilenamesArray, {}),
        (ar.BooleanArray, {}),
        (ar.Energies, {}),
        (ar.Bands, {}),
        (ar.VibrationalData, {"freq": []}),
        (ar.ScatteringData, {"freq": []}),
        (ar.ElectronicData, {"wavelen": []}),
        (ar.VibrationalActivities, {"freq": []}),
        (ar.ScatteringActivities, {"freq": []}),
        (ar.ElectronicActivities, {"wavelen": []}),
        (ar.Transitions, {}),
        (ar.Geometry, {"atoms": []}),
    ],
)
def test_empty_arrays(cls, kwargs):
    _ = cls(genre="", filenames=[], values=[], **kwargs)


@pytest.fixture
def get_repr_args_mock(monkeypatch):
    monkeypatch.setattr(
        ar.ArrayBase,
        "get_repr_args",
        mock.Mock(
            return_value={
                "genre": "bla",
                "filenames": ["f1", "f2", "f3"],
                "values": [1, 12, 15],
                "allow_data_inconsistency": False,
            }
        ),
    )
    return ar.ArrayBase.get_repr_args


@pytest.fixture
def clav(monkeypatch):
    monkeypatch.setattr(ar.dw, "calculate_average", mock.Mock(return_value=10))
    return ar.dw.calculate_average


@pytest.fixture
def lggr(monkeypatch):
    monkeypatch.setattr(ar, "logger", mock.Mock())
    return ar.logger


@pytest.fixture
def averagable():
    class Av(ar.FloatArray, ar.Averagable):
        associated_genres = ()

    return Av(genre="bla", filenames=["f1", "f2", "f3"], values=[3, 12, 15])


def test_average_conformers_energies_object(averagable, lggr, clav, get_repr_args_mock):
    en = mock.Mock(genre="foo", populations=[1, 1, 1])
    out = averagable.average_conformers(en)
    assert type(out) is type(averagable)
    assert out.values.tolist() == [10]
    assert clav.call_args[0][0].tolist() == [3, 12, 15]
    assert clav.call_args[0][1] == [1, 1, 1]
    get_repr_args_mock.assert_called()
    lggr.debug.assert_called_with("bla averaged by foo.")


def test_average_conformers_list(averagable, lggr, clav, get_repr_args_mock):
    out = averagable.average_conformers([1, 1, 1])
    assert type(out) is type(averagable)
    assert out.values.tolist() == [10]
    assert clav.call_args[0][0].tolist() == [3, 12, 15]
    assert clav.call_args[0][1].tolist() == [1, 1, 1]
    get_repr_args_mock.assert_called()
    lggr.debug.assert_called_with("bla averaged by unknown.")


@pytest.fixture
def filenames_mock(monkeypatch):
    mocked = mock.PropertyMock(return_value=["f1", "f2", "f3"])
    monkeypatch.setattr(ar.Energies, "filenames", mocked)
    return mocked


@pytest.fixture
def values_mock(monkeypatch):
    mocked = mock.PropertyMock(return_value=[3, 12, 15])
    monkeypatch.setattr(ar.Energies, "values", mocked)
    return mocked


@pytest.fixture
def convert_mock(monkeypatch):
    mocked = mock.PropertyMock(return_value=[3, 12, 15])
    monkeypatch.setattr(ar.Energies, "as_kcal_per_mol", mocked)
    return mocked


@pytest.fixture
def en():
    return ar.Energies(
        genre="bla", filenames=["f1", "f2", "f3"], values=[3, 12, 15], t=10
    )


def test_deltas(en, monkeypatch, filenames_mock, convert_mock):
    monkeypatch.setattr(ar.dw, "calculate_deltas", mock.Mock())
    _ = en.deltas
    convert_mock.assert_called()
    ar.dw.calculate_deltas.assert_called_with(en.as_kcal_per_mol)


def test_min_factors(en, monkeypatch, filenames_mock, convert_mock):
    monkeypatch.setattr(ar.dw, "calculate_min_factors", mock.Mock())
    _ = en.min_factors
    convert_mock.assert_called()
    ar.dw.calculate_min_factors.assert_called_with(en.as_kcal_per_mol, en.t)


def test_populations(en, monkeypatch, filenames_mock, convert_mock):
    monkeypatch.setattr(ar.dw, "calculate_populations", mock.Mock())
    _ = en.populations
    convert_mock.assert_called()
    ar.dw.calculate_populations.assert_called_with(en.as_kcal_per_mol, en.t)


def test_calculate_populations(en, monkeypatch, filenames_mock, convert_mock):
    monkeypatch.setattr(ar.dw, "calculate_populations", mock.Mock())
    en.calculate_populations(20)
    convert_mock.assert_called()
    ar.dw.calculate_populations.assert_called_with(en.as_kcal_per_mol, 20)


@pytest.fixture
def freq(monkeypatch):
    monkeypatch.setattr(
        ar.SpectralActivities,
        "frequencies",
        mock.PropertyMock(return_value=[[10, 20], [12, 21]]),
    )
    return ar.SpectralActivities.frequencies


@pytest.fixture
def vals(monkeypatch):
    monkeypatch.setattr(
        ar.SpectralActivities,
        "values",
        mock.PropertyMock(return_value=[[2, 5], [3, 7]]),
    )
    return ar.SpectralActivities.values


@pytest.fixture
def fnms(monkeypatch):
    monkeypatch.setattr(
        ar.SpectralActivities, "filenames", mock.PropertyMock(return_value=["f1", "f2"])
    )
    return ar.SpectralActivities.filenames


@pytest.fixture
def inten(monkeypatch):
    inten_mock = mock.MagicMock()
    inten_mock.converter = mock.Mock()
    inten_mock.__getitem__.side_effect = lambda _k: inten_mock.converter
    monkeypatch.setattr(ar.VibrationalActivities, "_intensities_converters", inten_mock)
    return inten_mock


@pytest.fixture
def inten_no_conversion(inten):
    def no_key(_k):
        raise KeyError

    inten.__getitem__.side_effect = no_key
    return inten


@pytest.fixture
def bars():
    return ar.VibrationalActivities("iri", [], [], [])


def test_intensieties(bars, inten, fnms, vals, freq):
    _ = bars.intensities
    inten.converter.assert_called_with(bars.values, bars.frequencies)


def test_intensieties_unsupported_genre(bars, inten_no_conversion, fnms, vals, freq):
    with pytest.raises(NotImplementedError):
        _ = bars.intensities


# test Geometry
@pytest.fixture
def geom():
    return ar.Geometry(
        genre="last_read_geom",
        filenames=["file1.out", "file2.out"],
        values=[
            [[30, 40, 60], [40, 60, 70], [50, 80, 10]],
            [[35, 45, 65], [45, 65, 75], [55, 85, 15]],
        ],
        atoms=[[1, 1, 1]],
    )


@pytest.fixture
def geom_incons():
    return ar.Geometry(
        genre="last_read_geom",
        filenames=["file1.out", "file2.out"],
        values=[
            [[30, 40, 60], [40, 60, 70], [50, 80, 10]],
            [[35, 45, 65], [45, 65, 75]],
        ],
        atoms=[[2, 2, 2], [2, 2]],
        allow_data_inconsistency=True,
    )


def test_atoms(geom):
    assert geom.atoms.tolist() == [[1, 1, 1]]


def test_atoms_as_symbols(geom, monkeypatch):
    monkeypatch.setattr(type(geom).atoms, "fsan", mock.Mock(return_value=[[6, 1, 8]]))
    geom.atoms = [["C", "H", "O"]]
    assert geom.atoms.tolist() == [[6, 1, 8]]


def test_atoms_two_dim(geom):
    geom.atoms = [[2, 2, 2], [2, 2, 2]]
    assert geom.atoms.tolist() == [[2, 2, 2]]


def test_atoms_two_dim_different(geom):
    with pytest.raises(InconsistentDataError):
        geom.atoms = [[2, 2, 2], [2, 1, 2]]


def test_atoms_two_dim_different_inconsistency_allowed(geom):
    geom.allow_data_inconsistency = True
    geom.atoms = [[2, 2, 2], [2, 1, 2]]
    assert geom.atoms.tolist() == [[2, 2, 2], [2, 1, 2]]


def test_molecule_values_varying_sizes_inconsistency_allowed(geom):
    geom.allow_data_inconsistency = True
    geom.values = [
        [[30, 40, 60], [40, 60, 70], [50, 80, 10]],
        [[35, 45, 65], [45, 65, 75]],
    ]
    np.testing.assert_array_equal(
        geom.values,
        [
            [[30, 40, 60], [40, 60, 70], [50, 80, 10]],
            [[35, 45, 65], [45, 65, 75], [0, 0, 0]],
        ],
    )


@pytest.mark.xfail(
    reason=(
        "Geometry.atoms.sanitizer currently " "does not support jagged nested lists."
    )
)
def test_atoms_varying_sizes_inconsistency_allowed(geom):
    geom.allow_data_inconsistency = True
    geom.atoms = [[2, 2, 2], [2, 2]]
    assert geom.atoms.tolist() == [[2, 2, 2], [2, 2, 0]]


def test_atoms_too_short(geom):
    with pytest.raises(ValueError):
        geom.atoms = [[2, 2]]


def test_atoms_not_matching_num_of_conformers(geom):
    with pytest.raises(ValueError):
        geom.atoms = [[2, 2, 2]] * 3


@pytest.fixture(scope="module")
def transitions_values():
    return [[[(11, 21, 0.5), (12, 22, -0.6)], [(31, 32, 0.9)]]]


@pytest.fixture
def transitions(transitions_values):
    return ar.Transitions(
        genre="transitions",
        filenames=["one"],
        values=transitions_values,
    )


def test_transitions_unpack(transitions_values):
    ground, excited, coefs = ar.Transitions.unpack_values(transitions_values)
    assert ground == [[[11, 12], [31]]]
    assert excited == [[[21, 22], [32]]]
    assert coefs == [[[0.5, -0.6], [0.9]]]


def test_transitions_highest(transitions):
    g, e, v, c = transitions.highest_contribution
    np.testing.assert_array_equal(v, [[-0.6, 0.9]])
    np.testing.assert_array_equal(c, [[2 * (-0.6) ** 2, 2 * 0.9 ** 2]])
    np.testing.assert_array_equal(e, [[22, 32]])
    np.testing.assert_array_equal(g, [[12, 31]])


def test_all_define_full_name(any_genre):
    assert any_genre in _ARRAY_CONSTRUCTORS[any_genre]._full_name_ref


def test_all_activities_calculates_spectra(any_activity):
    def fitting_func(i, f, a, w):
        return a * w  # appropriate shape, values doesn't matter

    constructor = _ARRAY_CONSTRUCTORS[any_activity]
    array = constructor(any_activity, ["one", "two"], [[1], [2]], [[90], [110]])
    spc = array.calculate_spectra(0, 200, 1, 1, fitting_func)
    assert spc.genre



import re
from unittest.mock import Mock, NonCallableMock, patch

import pytest

from tesliper.exceptions import InvalidStateError
from tesliper.extraction import ParserBase


@pytest.fixture
def subclass_params():
    return {
        "initial": Mock(spec=[]),
        "parse": Mock(spec=[]),
        "purpose": "",
        "extensions": tuple(),
    }


@pytest.fixture
def params_with_state(subclass_params, method):
    method.is_state = True
    subclass_params["mystate"] = method
    return subclass_params


@pytest.fixture
def prs(subclass_params):
    cls = type("Prsr", (ParserBase,), subclass_params)
    return cls()


@pytest.fixture
def prs_with_state(params_with_state):
    cls = type("Prsr", (ParserBase,), params_with_state)
    return cls()


@pytest.fixture
def method():
    return Mock(spec=[])


def test_instantiated_directly():
    with pytest.raises(TypeError):
        ParserBase()


def test_state_no_arguments():
    with pytest.raises(TypeError):
        ParserBase.state()


def test_state_not_callable():
    func = NonCallableMock()
    with pytest.raises(InvalidStateError):
        ParserBase.state(func)


def test_state_decorator(method):
    func = ParserBase.state(method)
    assert hasattr(func, "is_state")
    assert func.is_state


def test_state_decorator_with_trigger(method):
    func = ParserBase.state(method, "mytrigger")
    assert hasattr(func, "is_state")
    assert func.is_state
    assert hasattr(func, "trigger")
    assert func.trigger == "mytrigger"


def test_subclassed(prs):
    assert "initial" in prs.states
    assert prs.initial == prs.workhorse


def test_subclassed_with_state(prs_with_state):
    assert hasattr(prs_with_state, "mystate")
    assert "mystate" in prs_with_state.states
    assert prs_with_state.initial == prs_with_state.workhorse


def test_subclassed_with_trigger():
    class Subclass(ParserBase):
        purpose = ""
        extensions = tuple()

        def parse(self, lines):
            pass

        def initial(self, line):
            pass

        @ParserBase.state(trigger="trigger")
        def mystate(prs, method):
            pass

    with patch("re.compile") as mocked_re_compile:
        mocked_re_compile.return_value = "compiled_trigger"
        obj = Subclass()
    assert hasattr(obj, "mystate")
    assert hasattr(obj.mystate, "is_state")
    assert "mystate" in obj.triggers
    assert mocked_re_compile.called
    assert obj.triggers["mystate"] == "compiled_trigger"


def test_add_state(prs, method):
    func = method
    func.__name__ = "mystate"
    prs.add_state(func)
    assert "mystate" in prs.states


def test_remove_state(prs_with_state):
    assert "mystate" in prs_with_state.states
    prs_with_state.remove_state("mystate")
    assert "mystate" not in prs_with_state.states


def test_add_state_with_trigger_as_string(prs, method):
    func = method
    func.__name__ = "mystate"
    func.trigger = "regex"
    prs.add_state(func)
    assert "mystate" in prs.states
    assert "mystate" in prs.triggers
    assert prs.triggers["mystate"] == re.compile("regex")


def test_add_state_with_trigger_as_regex(prs, method):
    func = method
    func.__name__ = "mystate"
    func.trigger = re.compile("regex")
    prs.add_state(func)
    assert "mystate" in prs.states
    assert "mystate" in prs.triggers
    assert prs.triggers["mystate"] == re.compile("regex")


def test_remove_state_with_trigger(params_with_state):
    params_with_state["mystate"].trigger = "regex"
    cls = type("Prsr", (ParserBase,), params_with_state)
    obj = cls()
    assert "mystate" in obj.states
    assert "mystate" in obj.triggers
    obj.remove_state("mystate")
    assert "mystate" not in obj.states
    assert "mystate" not in obj.triggers


def test_remove_invalid(prs):
    with pytest.raises(InvalidStateError):
        prs.remove_state("invalid")


def test_set_workhorse(prs, method):
    func = method
    prs.workhorse = func
    assert prs.workhorse is func


def test_set_workhorse_name(prs_with_state):
    prs_with_state.workhorse = "mystate"
    assert prs_with_state.workhorse is prs_with_state.mystate


def test_set_workhorse_invalid(prs):
    def func():
        prs.workhorse = "invalid"

    with pytest.raises(InvalidStateError):
        func()


def test_initial_matching(prs, method):
    prs.initial.side_effect = lambda line: super(type(prs), prs).initial(line)
    prs.states["mystate"] = method
    prs.triggers["mystate"] = re.compile("yes")
    out = prs.initial("yes, this line will match")
    prs.initial.assert_called()
    assert out == {}
    assert prs.workhorse is method


def test_initial_not_matching(prs, method):
    prs.initial.side_effect = lambda line: super(type(prs), prs).initial(line)
    prs.states["mystate"] = method
    prs.triggers["mystate"] = re.compile("yes")
    out = prs.initial("this line will not match")
    prs.initial.assert_called()
    assert out == {}
    assert prs.workhorse is prs.initial


def test_parse(prs):
    prs.initial.return_value = {"example": "data"}
    prs.parse.side_effect = lambda lines: super(type(prs), prs).parse(lines)
    out = prs.parse(["some line"])
    prs.initial.assert_called()
    assert out == {"example": "data"}


def test_parse_initial_resetting(prs, method):
    method.return_value = {}
    prs.workhorse = method
    prs.parse.side_effect = lambda lines: super(type(prs), prs).parse(lines)
    prs.parse(["some line"])
    method.assert_called()
    assert prs.workhorse is prs.initial


def test_parse_raises_not_iterable(prs, method):
    method.return_value = 1
    prs.workhorse = method
    prs.parse.side_effect = lambda lines: super(type(prs), prs).parse(lines)
    with pytest.raises(InvalidStateError):
        prs.parse(["some line"])
    assert prs.workhorse is prs.initial


def test_parse_raises_cant_convert_value(prs, method):
    method.return_value = "wrong sequence"
    prs.workhorse = method
    prs.parse.side_effect = lambda lines: super(type(prs), prs).parse(lines)
    with pytest.raises(InvalidStateError):
        prs.parse(["some line"])
    assert prs.workhorse is prs.initial



from pathlib import Path

import pytest
from hypothesis import assume, given
from hypothesis import strategies as st

from tesliper.writing.gjf_writer import GjfWriter, _format_coordinates


@given(
    st.lists(st.integers(min_value=1, max_value=118), min_size=4, max_size=4),
    st.lists(
        st.floats(min_value=-100, max_value=100, exclude_min=True, exclude_max=True),
        min_size=12,
        max_size=12,
    ),
)
def test__format_coordinates(atoms, coords):
    coords = [coords[0 + 3 * n : 3 + 3 * n] for n in range(len(atoms))]
    output = list(_format_coordinates(coords, atoms))
    assert len(output) == len(atoms)
    assert all([len(output[0]) == len(output[n]) for n in range(len(output))])


def test_writer_init():
    gjfwriter = GjfWriter(destination="")
    assert gjfwriter.destination == Path("")
    assert gjfwriter.mode == "x"
    assert gjfwriter.link0 == {}
    assert gjfwriter.route == "#"
    assert gjfwriter.comment == "No information provided."
    assert gjfwriter.post_spec == ""


@given(st.text())
def test_link0_unknown(key):
    assume(key.lower() not in GjfWriter._link0_commands)
    with pytest.raises(ValueError):
        GjfWriter(destination="", link0={key: ""})


@pytest.mark.xfail(reason="To be created")
@given(
    st.dictionaries(
        st.one_of(*[st.just(k) for k in GjfWriter._link0_commands]), st.text()
    )
)
def test_link0_known(dict):
    pytest.fail("To be created")


@given(st.text())
def test_route_str_not_startswith_hash(commands):
    assume(not commands.strip().startswith("#"))
    gjfwriter = GjfWriter(destination="", route=commands)
    assert gjfwriter.route == " ".join(["#"] + commands.split())
    assert gjfwriter._route == ["#"] + commands.split()


@given(st.builds(lambda s: "#" + s, st.text()))
def test_route_str_startswith_hash(commands):
    gjfwriter = GjfWriter(destination="", route=commands)
    assert gjfwriter.route == " ".join(commands.split())
    assert gjfwriter._route == commands.split()


st_not_st = st.one_of(
    st.none(),
    st.integers(),
    st.floats(),
    st.sets(st.text(), min_size=1),
    st.dictionaries(st.text(), st.text(), min_size=1),
)


@given(st_not_st)
def test_route_wrong_type(commands):
    with pytest.raises(TypeError):
        GjfWriter(destination="", route=commands)


@given(st.lists(st_not_st, min_size=1))
def test_route_sequence_wrong_type(commands):
    with pytest.raises(TypeError):
        GjfWriter(destination="", route=commands)


def test_route_empty():
    gjfwriter = GjfWriter(destination="", route="")
    assert gjfwriter.route == "#"
    assert gjfwriter._route == ["#"]



import logging
import tkinter as tk
from tkinter import ttk

logger = logging.getLogger(__name__)


class HintedEntry(ttk.Entry):
    """Entry that displays a hint and changes style when empty."""

    def __init__(
        self, parent, hint="Enter text here", hinted_style="Hinted.TEntry", **kwargs
    ):
        ttk.Style().configure("Hinted.TEntry", foreground="grey")
        self._hint = hint
        self.variable = kwargs["textvariable"] = kwargs.get(
            "textvariable", tk.StringVar()
        )
        self._style = kwargs.get("style", "TEntry")
        self._hinted_style = hinted_style
        kwargs["style"] = self._style if self.variable.get() else self._hinted_style
        super().__init__(parent, **kwargs)
        if not self.variable.get():
            self.variable.set(hint)
        self.bind("<FocusIn>", self.entry_focus_in)
        self.bind("<FocusOut>", self.entry_focus_out)

    def entry_focus_in(self, _event=None):
        normal_state = str(self.cget("state")) == "normal"
        if self.is_hinted() and normal_state:
            self.variable.set("")
            super().configure(style=self._style)

    def entry_focus_out(self, _event=None):
        if not self.variable.get():
            self.variable.set(self._hint)
            super().configure(style=self._hinted_style)

    def get(self):
        return "" if self.is_empty() else self.variable.get()

    def set(self, value):
        is_normal = value or self.focus_get() is self
        self.variable.set(value if is_normal else self._hint)
        logger.debug(f" Entry is {'normal' if is_normal else 'hinted'}.")
        super().configure(style=self._style if is_normal else self._hinted_style)

    def is_empty(self):
        return not self.variable.get() or self.is_hinted()

    def is_hinted(self):
        return self.cget("style") == self._hinted_style

    def configure(self, cnf=None, **kwargs):
        is_hinted = self.is_hinted()
        hint = kwargs.pop("hint", None)
        if hint is not None:
            self._hint = hint
            if is_hinted:
                self.variable.set(hint)
        style = kwargs.pop("style", None)
        if style is not None:
            self._style = style
        hinted_style = kwargs.pop("hinted_style", None)
        if hinted_style is not None:
            self._hinted_style = hinted_style
        if style or hinted_style:
            kwargs["style"] = self._hinted_style if is_hinted else self._style
        super().configure(cnf, **kwargs)



# IMPORTS
import logging as lgg
import tkinter as tk
import tkinter.ttk as ttk
from collections import OrderedDict

import numpy as np

import tesliper

# LOGGER
logger = lgg.getLogger(__name__)


# CLASSES
class BoxVar(tk.BooleanVar):
    def __init__(self, box, *args, **kwargs):
        self.box = box
        if "value" not in kwargs:
            kwargs["value"] = True
        super().__init__(*args, **kwargs)

    @property
    def tesliper(self):
        return self.box.winfo_toplevel().tesliper

    def _set(self, value):
        super().set(value)
        tags = () if value else "discarded"
        self.box.tree.item(self.box.index, tags=tags)

    def set(self, value):
        # set is not called by tkinter when checkbutton is clicked
        self.tesliper.conformers.kept[int(self.box.index)] = bool(value)
        for tree in self.box.tree.trees.values():
            try:
                tree.boxes[self.box.index].var._set(value)
            except KeyError:
                logger.debug(f"{tree} doesn't have box iid {self.box.index}")


class Checkbox(ttk.Checkbutton):
    def __init__(self, master, tree, index, *args, **kwargs):
        self.frame = ttk.Frame(master, width=17, height=20)
        self.tree = tree
        self.index = index
        self.var = BoxVar(self)
        kwargs["variable"] = self.var
        super().__init__(self.frame, *args, command=self.clicked, **kwargs)
        self.frame.pack_propagate(False)
        self.frame.grid_propagate(False)
        self.grid(column=0, row=0)

    def clicked(self):
        logger.debug(f"box index: {self.index}")
        value = self.var.get()
        self.var.set(value)
        self.event_generate("<<KeptChanged>>")


class CheckTree(ttk.Treeview):
    trees = dict()

    def __init__(self, master, name, **kwargs):
        CheckTree.trees[name] = self
        self.frame = ttk.Frame(master)
        style = ttk.Style()
        style.layout(
            "borderless.Treeview", [("mystyle.Treeview.treearea", {"sticky": "nswe"})]
        )
        super().__init__(self.frame, **kwargs, style="borderless.Treeview")
        self.grid(column=0, row=0, rowspan=2, columnspan=2, sticky="nwse")
        tk.Grid.columnconfigure(self.frame, 1, weight=1)
        tk.Grid.rowconfigure(self.frame, 1, weight=1)
        self.vsb = ttk.Scrollbar(self.frame, orient="vertical", command=self.on_bar)
        self.vsb.grid(column=2, row=0, rowspan=2, sticky="nse")

        self.tag_configure("discarded", foreground="gray")

        # Sort button
        but_frame = ttk.Frame(self.frame, height=24, width=17)
        but_frame.grid(column=0, row=0)
        but_frame.grid_propagate(False)
        tk.Grid.columnconfigure(but_frame, 0, weight=1)
        tk.Grid.rowconfigure(but_frame, 0, weight=1)
        style.configure(
            "sorting.TButton", borderwidth=5, highlightthickness=1, relief="flat"
        )
        self.but_sort = ttk.Button(
            but_frame, style="sorting.TButton", command=self._sort_button
        )
        self.but_sort.grid(column=0, row=0, sticky="nwes")

        # Boxes
        self.canvas = tk.Canvas(
            self.frame,
            width=17,
            borderwidth=0,
            background="#ffffff",
            highlightthickness=0,
        )
        self.canvas.configure(yscrollcommand=self.vsb.set)
        self.canvas.grid(column=0, row=1, sticky="ns")
        self.boxes_frame = ttk.Frame(self.canvas)
        self.canvas.create_window(
            (0, 0), window=self.boxes_frame, anchor="nw", tags="boxes_frame"
        )

        self.boxes_frame.bind("<Configure>", self.onFrameConfigure)
        self.configure(yscrollcommand=self.yscroll)
        self.boxes = OrderedDict()

        self.owned_children = OrderedDict()
        self.children_names = OrderedDict()
        self.winfo_toplevel().bind("<<Clear>>", CheckTree.clear, "+")

    @property
    def tesliper(self):
        return self.winfo_toplevel().tesliper

    @property
    def blade(self):
        return [box.var.get() for box in self.boxes.values()]

    @property
    def dummy(self):
        ls = [self.item(i)["text"] for i in self.get_children()]
        ls = sorted(ls)
        dummy = tesliper.dw.Data("dummy", filenames=ls)
        dummy.trimmer.set(self.blade)
        return dummy

    def _sort_button(self, reverse=True):
        ls = [(b.var.get(), b.index) for b in self.boxes.values()]
        ls.sort(reverse=reverse)
        for i, (val, iid) in enumerate(ls):
            box = self.boxes[iid]
            self.move(iid, "", i)
            box.frame.grid_forget()
            box.frame.grid_propagate(False)
            box.frame.grid(column=0, row=i, sticky="n", pady=0)
        self.but_sort.configure(command=lambda: self._sort_button(not reverse))

    def _sort(self, col, reverse=True):
        # empty records always at the end
        empty = float("-inf") if reverse else float("inf")
        try:
            ls = [(self.set(iid, col), iid) for iid in self.get_children("")]
        except tk.TclError:
            ls = [(self.item(iid)["text"], iid) for iid in self.get_children("")]
        try:
            ls = [(empty if v == "--" else float(v), iid) for v, iid in ls]
        except ValueError:
            pass
        ls.sort(reverse=reverse)
        for i, (val, iid) in enumerate(ls):
            self.move(iid, "", i)
            box = self.boxes[iid]
            box.frame.grid_forget()
            box.frame.grid_propagate(False)
            box.frame.grid(column=0, row=i, sticky="n", pady=0)
        self.heading(col, command=lambda: self._sort(col, not reverse))

    def heading(self, col, *args, command=None, **kwargs):
        command = command if command is not None else lambda: self._sort(col)
        return super().heading(col, *args, command=command, **kwargs)

    def onFrameConfigure(self, event):
        """Reset the scroll region to encompass the inner frame"""
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))

    def yscroll(self, *args):
        logger.debug(args)
        self.canvas.yview_moveto(args[0])
        logger.debug(self.canvas.yview())

    def _insert(self, parent="", index=tk.END, iid=None, **kw):
        logger.debug(
            f"CALLED on {self.__class__} with parameters: parent={parent!r}, "
            f"index={index!r}, iid={iid!r}, kw={kw}"
        )
        if kw["text"] in self.owned_children:
            return iid
        box = Checkbox(self.boxes_frame, self, index=iid)
        box.frame.grid(column=0, row=len(self.boxes))
        self.boxes[box.index] = box
        iid = super().insert(parent, index, iid=str(box.index), **kw)
        self.owned_children[kw["text"]] = iid
        self.children_names[iid] = kw["text"]
        return iid

    def insert(self, parent="", index=tk.END, iid=None, **kw):
        try:
            text = kw["text"]
        except KeyError:
            raise TypeError("Required keyword argument 'text' not found.")
        if iid is not None:
            logger.debug("Overriding passed iid value.")
        if text in self.trees["main"].owned_children:
            iid = self.trees["main"].owned_children[text]
        else:
            iid = str(self.trees["main"].curr_iid)
            self.trees["main"].curr_iid += 1
        for tree in CheckTree.trees.values():
            tree._insert(parent=parent, index=index, iid=iid, **kw)

    def delete(self, *items):
        raise NotImplementedError(
            "Removing individual items not supported. "
            "To clear all trees, use `CheckTree.clear()`."
        )

    @classmethod
    def clear(cls, _event=None):
        for tree in CheckTree.trees.values():
            items = tree.get_children()
            super(CheckTree, tree).delete(*items)
            for box in tree.boxes.values():
                box.destroy()
            tree.boxes = OrderedDict()
            tree.owned_children = OrderedDict()
            tree.children_names = OrderedDict()
            cls.trees["main"].curr_iid = 0

    def on_bar(self, *args):
        self.yview(*args)
        # logger.debug(args)
        # logger.debug(self.canvas.yview())

    def refresh(self):
        pass
        # logger.debug(f"Called .refresh on {type(self)}")
        # kept = self.tslr.conformers.kept
        # boxes = self.boxes
        # for iid, name in self.children_names.items():
        #     boxes[iid].var.set(kept[int(iid)])


class EnergiesView(CheckTree):
    formats = dict(
        values=lambda v: "{:.6f}".format(v),
        deltas=lambda v: "{:.4f}".format(v),
        min_factors=lambda v: "{:.4f}".format(v),
        populations=lambda v: "{:.4f}".format(v * 100),
    )
    e_keys = "ten ent gib scf zpe".split(" ")

    def __init__(self, parent, **kwargs):
        kwargs["columns"] = self.e_keys
        super().__init__(parent, "energies", **kwargs)

        # Columns
        for cid, text in zip(
            ["#0"] + self.e_keys,
            "Filenames Thermal Enthalpy Gibbs SCF Zero-Point".split(" "),
        ):
            if not cid == "#0":
                self.column(cid, width=100, anchor="e", stretch=False)
            self.heading(cid, text=text)
        self.column("#0", width=150)

    def refresh(self, show):
        logger.debug("Going to update by showing {}.".format(show))
        if show == "values":
            # we don't want to hide energy values of non-kept conformer
            with self.tesliper.conformers.untrimmed:
                scope = self.tesliper.energies
        else:
            scope = self.tesliper.energies
        formatter = self.formats[show]
        # conformers are always in the same order, so we can use iterator for values
        # and only request next() when conformer's name is known by genre's DataArray
        values = {key: iter(getattr(scope[key], show)) for key in self.e_keys}
        fnames = {key: set(scope[key].filenames) for key in self.e_keys}
        for name, iid in self.owned_children.items():
            # owned_children is OrderedDict, so we get name and iid in ordered
            # they were inserted to treeview, which is same as order of data
            # stored in Tesliper instance
            for key in self.e_keys:
                # if this conformer's kept value is False,
                # use -- in place of missing values
                value = formatter(next(values[key])) if name in fnames[key] else "--"
                self.set(iid, column=key, value=value)


class ConformersOverview(CheckTree):
    def __init__(self, master, **kwargs):
        kwargs["columns"] = "term opt en ir vcd uv ecd ram roa " "imag stoich".split(
            " "
        )
        super().__init__(master, "main", **kwargs)
        self.curr_iid = 0

        # Columns
        self.column("#0", width=150)
        self.heading("#0", text="Filenames")
        for cid, text in zip(
            "term opt en ir vcd uv ecd ram roa imag stoich".split(" "),
            "Termination Opt Energy IR VCD UV ECD Raman ROA Imag "
            "Stoichiometry".split(" "),
        ):
            width = (
                80
                if cid == "term"
                else 90
                if cid == "stoich"
                else 50
                if cid in ("en", "ram")
                else 35
            )
            self.column(cid, width=width, anchor="center", stretch=False)
            self.heading(cid, text=text)
        self.__max_length = 0

    def _insert(self, parent="", index=tk.END, iid=None, **kw):
        # TO DO: correct wrong files counting when smaller set is extracted
        # first
        text = kw["text"]
        conf = self.tesliper.conformers[text]
        values = {
            "term": "normal" if conf["normal_termination"] else "ERROR",
            "opt": "n/a"
            if "optimization_completed" not in conf
            else "ok"
            if conf["optimization_completed"]
            else False,
            "en": "ok" if all(e in conf for e in EnergiesView.e_keys) else False,
            "ir": "ok" if "dip" in conf else False,
            "vcd": "ok" if "rot" in conf else False,
            "uv": "ok" if "vosc" in conf else False,
            "ecd": "ok" if "vrot" in conf else False,
            "ram": "ok" if "raman1" in conf else False,
            "roa": "ok" if "roa1" in conf else False,
        }
        if "freq" in conf:
            freqs = self.tesliper.conformers[text]["freq"]
            imag = str((np.array(freqs) < 0).sum())
            values["imag"] = imag
        else:
            values["imag"] = False
        try:
            values["stoich"] = conf["stoichiometry"]
        except KeyError:
            values["stoich"] = "--"
        iid = super()._insert(parent=parent, index=index, iid=iid, **kw)
        for k, v in values.items():
            self.set(iid, k, v or "X")
        return iid

    @classmethod
    def test_populate(cls, master, num=30):
        import random
        import string

        new = cls(master, columns=("b"))
        new.heading("b", text="afasdgf")
        new.heading("#0", text="asdgasdfg")
        gen = ("".join(random.choices(string.ascii_lowercase, k=7)) for x in range(num))
        for x, bla in enumerate(gen):
            new.insert(text=bla + " " + str(x), values=[x])
        return new



"""Serialization and deserialization of :class:`.Tesliper` objects."""
import json
import logging
import zipfile
from json.decoder import JSONArray
from json.scanner import py_make_scanner
from pathlib import Path
from typing import Any, Dict, Iterable, List, Union

import tesliper  # absolute import to solve problem of circular imports
from tesliper import datawork as dw
from tesliper.glassware import Conformers, SingleSpectrum, Spectra

logger = logging.getLogger(__name__)


class ArchiveWriter:
    """Class for serialization of Tesliper objects.

    Structure of the produced archive::

        .
        ├───arguments: {input_dir=str, output_dir=str, wanted_files=[str]}
        ├───parameters: {"ir": {params}, ..., "roa": {params}}
        ├───conformers
        │   ├───arguments: {"allow_data_inconsistency": bool,
        │   │               "temperature_of_the_system": float}
        │   ├───filenames: [str]
        │   ├───kept: [bool]
        │   └───data
        │       ├───filename_1: {genre=str: data}
        |       ...
        │       └───filename_N: {genre=str: data}
        └───spectra
            ├───experimental
            │   ├───spectra_genre_1: {attr_name: SingleSpectrum.attr}
            |   ...
            │   └───spectra_genre_N: {attr_name: SingleSpectrum.attr}
            ├───calculated
            │   ├───spectra_genre_1: {attr_name: Spectra.attr}
            |   ...
            │   └───spectra_genre_N: {attr_name: Spectra.attr}
            └───averaged
                ├───spectra_genre_1-energies-genre-1: {attr_name: SingleSpectrum.attr}
                ...
                └───spectra_genre_N-energies-genre-N: {attr_name: SingleSpectrum.attr}
    """

    def __init__(
        self, destination: Union[str, Path], mode: str = "x", encoding: str = "utf-8"
    ):
        """
        Parameters
        ----------
        destination : Union[str, Path]
            Path to target file.
        mode : str, optional
            Specifies how writing to file should be handled. Should be one of
            characters: 'a' (append to existing file), 'x' (only write if file doesn't
            exist yet), or 'w' (overwrite file if it already exists). Defaults to "x".
        encoding : str, optional
            Encoding of the output, by default "utf-8"
        """
        self.mode = mode
        self.destination = destination
        self.encoding = encoding
        self.root = None

    def __enter__(self):
        return self.open()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @property
    def mode(self):
        """Specifies how writing to file should be handled. Should be one of characters:
        "a", "x", or "w".
        "a" - append to existing file;
        "x" - only write if file doesn't exist yet;
        "w" - overwrite file if it already exists.

        Raises
        ------
        ValueError
            If given anything other than "a", "x", or "w".
        """
        return self._mode

    @mode.setter
    def mode(self, mode):
        if mode not in ("a", "x", "w"):
            raise ValueError("Mode should be 'a', 'x', or 'w'.")
        self._mode = mode

    @property
    def destination(self) -> Path:
        """pathlib.Path: Directory, to which generated files should be written.

        Raises
        ------
        FileNotFoundError
            If given destination doesn't exist or is not a directory.
        """
        return vars(self)["destination"]

    @destination.setter
    def destination(self, destination: Union[str, Path]) -> None:
        destination = Path(destination)
        if not destination.exists() and self.mode == "a":
            raise FileNotFoundError(
                "Mode 'a' was specified, but given file doesn't exist."
            )
        elif destination.exists() and self.mode == "x":
            raise FileExistsError(
                "Mode 'x' was specified, but given file already exists."
            )
        elif not destination.parent.exists():
            raise FileNotFoundError("Parent directory of specified file doesn't exist.")
        else:
            logger.debug(f"File {destination} ok for writing.")
        vars(self)["destination"] = destination

    def open(self):
        self.root = zipfile.ZipFile(self.destination, mode=self.mode)
        return self

    def close(self):
        self.root.close()

    def write(self, obj: "tesliper.Tesliper"):
        with self:
            self._write_arguments(
                obj.input_dir, obj.output_dir, obj.wanted_files, obj.quantum_software
            )
            self._write_parameters(obj.parameters)
            self._write_conformers(obj.conformers)
            for spc in obj.averaged.values():
                self._write_averaged(spc)
            for spc in obj.spectra.values():
                self._write_calculated(spc)
            for spc in obj.experimental.values():
                self._write_experimental(spc)

    def _write_arguments(
        self,
        input_dir: Union[Path, str] = None,
        output_dir: Union[Path, str] = None,
        wanted_files: Iterable[str] = None,
        quantum_software: str = None,
    ):
        with self.root.open("arguments.json", mode="w") as handle:
            handle.write(
                self.jsonencode(
                    {
                        "input_dir": str(input_dir) if input_dir else None,
                        "output_dir": str(output_dir) if output_dir else None,
                        "wanted_files": list(wanted_files) if wanted_files else None,
                        "quantum_software": quantum_software or None,
                    }
                )
            )

    def _write_parameters(self, parameters: dict):
        # TODO: Implement more universal way of serializing fitting
        #       this won't deserialize custom fitting functions
        to_write = {key: params.copy() for key, params in parameters.items()}
        for params in to_write.values():
            params["fitting"] = params["fitting"].__name__
        with self.root.open("parameters.json", mode="w") as handle:
            handle.write(self.jsonencode(to_write))

    def _write_conformers(self, conformers: Conformers):
        self._write_conformers_arguments(
            allow_data_inconsistency=conformers.allow_data_inconsistency,
            temperature_of_the_system=conformers.temperature,
        )
        self._write_filenames(conformers.filenames)
        self._write_kept(conformers.kept)
        for filename in conformers.filenames:
            self._write_mol(filename=filename, mol=conformers[filename])

    def _write_conformers_arguments(self, **kwargs):
        with self.root.open("conformers/arguments.json", mode="w") as handle:
            handle.write(self.jsonencode(kwargs))

    def _write_filenames(self, filenames: List[str]):
        with self.root.open("conformers/filenames.json", mode="w") as handle:
            handle.write(self.jsonencode(filenames))

    def _write_mol(self, filename: str, mol: dict):
        with self.root.open(f"conformers/data/{filename}.json", mode="w") as handle:
            handle.write(self.jsonencode(mol))

    def _write_kept(self, kept: List[bool]):
        with self.root.open("conformers/kept.json", mode="w") as handle:
            handle.write(self.jsonencode(kept))

    def _write_experimental(self, spectrum: SingleSpectrum):
        path = f"spectra/experimental/{spectrum.genre}.json"
        with self.root.open(path, mode="w") as handle:
            handle.write(
                self.jsonencode(
                    {
                        "genre": spectrum.genre,
                        "filenames": spectrum.filenames.tolist(),
                        "values": spectrum.values.tolist(),
                        "abscissa": spectrum.abscissa.tolist(),
                        "width": spectrum.width,
                        "fitting": spectrum.fitting,
                        "scaling": spectrum.scaling,
                        "offset": spectrum.offset,
                        "averaged_by": spectrum.averaged_by,
                    }
                )
            )

    def _write_calculated(self, spectra: Spectra):
        path = f"spectra/calculated/{spectra.genre}.json"
        with self.root.open(path, mode="w") as handle:
            handle.write(
                self.jsonencode(
                    {
                        "genre": spectra.genre,
                        "filenames": spectra.filenames.tolist(),
                        "values": spectra.values.tolist(),
                        "abscissa": spectra.abscissa.tolist(),
                        "width": spectra.width,
                        "fitting": spectra.fitting,
                        "scaling": spectra.scaling,
                        "offset": spectra.offset,
                        "allow_data_inconsistency": spectra.allow_data_inconsistency,
                    }
                )
            )

    def _write_averaged(self, spectrum: SingleSpectrum):
        path = f"spectra/averaged/{spectrum.genre}-{spectrum.averaged_by}.json"
        with self.root.open(path, mode="w") as handle:
            handle.write(
                self.jsonencode(
                    {
                        "genre": spectrum.genre,
                        "filenames": spectrum.filenames.tolist(),
                        "values": spectrum.values.tolist(),
                        "abscissa": spectrum.abscissa.tolist(),
                        "width": spectrum.width,
                        "fitting": spectrum.fitting,
                        "scaling": spectrum.scaling,
                        "offset": spectrum.offset,
                        "averaged_by": spectrum.averaged_by,
                    }
                )
            )

    def jsonencode(
        self,
        obj: Any,
        *,
        skipkeys=False,
        ensure_ascii=True,
        check_circular=True,
        allow_nan=True,
        cls=None,
        indent=None,
        separators=None,
        default=None,
        sort_keys=False,
        **kw,
    ) -> bytes:
        """json.dumps wrapper, that encodes JSON produced."""
        return json.dumps(
            obj,
            skipkeys=skipkeys,
            ensure_ascii=ensure_ascii,
            check_circular=check_circular,
            allow_nan=allow_nan,
            cls=cls,
            indent=indent,
            separators=separators,
            default=default,
            sort_keys=sort_keys,
            **kw,
        ).encode(self.encoding)


class ArchiveLoader:
    """Class for deserialization of Tesliper objects."""

    def __init__(self, source: Union[str, Path], encoding: str = "utf-8"):
        """
        Parameters
        ----------
        source : Union[str, Path]
            Path to the source file.
        encoding : str, optional
            Source file encoding, by default "utf-8".
        """
        self.source = source
        self.encoding = encoding
        self.root = None

    def __enter__(self):
        return self.open()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def open(self):
        self.root = zipfile.ZipFile(self.source, mode="r")
        return self

    def close(self):
        self.root.close()

    @property
    def source(self) -> Path:
        """pathlib.Path: File, from which data should read.

        Notes
        -----
        If str given, it will be converted to pathlib.Path.

        Raises
        ------
        FileNotFoundError
            If given destination doesn't exist.
        """
        return self._destination

    @source.setter
    def source(self, destination: Union[str, Path]) -> None:
        destination = Path(destination)
        if not destination.exists():
            raise FileNotFoundError("Given destination doesn't exist.")
        self._destination = destination

    def load(self) -> "tesliper.Tesliper":
        with self:
            tslr = tesliper.Tesliper(**self._load("arguments.json"))
            tslr.parameters = self._load_parameters()
            filenames = self._load("conformers/filenames.json")
            mols = (
                (
                    name,
                    self.jsondecode(self.root.read(f"conformers/data/{name}.json")),
                )
                for name in filenames
            )  # iterator producing key-value pairs
            tslr.conformers = Conformers(
                mols, **self._load("conformers/arguments.json")
            )
            tslr.conformers.kept = self._load("conformers/kept.json")
            for file in self.root.namelist():
                if "experimental" in file:
                    params = self._load(file)
                    tslr.experimental[params["genre"]] = SingleSpectrum(**params)
                elif "calculated" in file:
                    params = self._load(file)
                    tslr.spectra[params["genre"]] = Spectra(**params)
                elif "averaged" in file:
                    params = self._load(file)
                    tslr.averaged[
                        (params["genre"], params["averaged_by"])
                    ] = SingleSpectrum(**params)
        return tslr

    def _load(self, dest):
        return self.jsondecode(self.root.read(dest))

    def _load_parameters(self):
        parameters = self._load("parameters.json")
        for params in parameters.values():
            params["fitting"] = getattr(dw, params["fitting"])
        return parameters

    def jsondecode(
        self,
        string: bytes,
        *,
        cls=None,
        object_hook=None,
        parse_float=None,
        parse_int=None,
        parse_constant=None,
        object_pairs_hook=None,
        **kw,
    ) -> Any:
        """json.loads wrapper, that decodes bytes before parsing as JSON."""
        return json.loads(
            string.decode(self.encoding),
            cls=cls,
            object_hook=object_hook,
            parse_float=parse_float,
            parse_int=parse_int,
            parse_constant=parse_constant,
            object_pairs_hook=object_pairs_hook,
            **kw,
        )



"""Objects for data serialization.

Aside from concrete implementations of :class:`.WriterBase`-derived classes for
particular file formats, this module provides a :func:`.writer` factory function that
allows to dynamically retrieve particular writer objects. This function is used by
:class:`.Tesliper` when exporting data to allow for use of user-provided
:class:`.WriterBase` subclasses.
"""
from .csv_writer import CsvWriter
from .gjf_writer import GjfWriter
from .serializer import ArchiveLoader, ArchiveWriter
from .txt_writer import TxtWriter
from .writer_base import WriterBase, writer
from .xlsx_writer import XlsxWriter



"""Parser for spectra files."""
import csv
import logging as lgg
from pathlib import Path
from typing import Optional, Union

import numpy as np

from .parser_base import ParserBase

logger = lgg.getLogger(__name__)


class SpectraParser(ParserBase):
    """Parser for files containing spectral data. It can parse .txt (in "x y" format)
    and .csv files, returning an numpy.ndarray with loaded spectrum. Parsing process
    may be customized by specifying what delimiter of values should be expected
    and in which column x- and y-values are, if there are more than 2 columns of data.
    If file contains any header, it is ignored.
    """

    purpose = "spectra"
    extensions = (".txt", ".xy", ".csv")

    def __init__(self):
        super().__init__()
        self.delimiter = None
        self.xcolumn = 0
        self.ycolumn = 1

    def parse(
        self,
        filename: Union[str, Path],
        delimiter: Optional[str] = None,
        xcolumn: int = 0,
        ycolumn: int = 1,
    ) -> np.ndarray:
        """Loads spectral data from file to numpy.array. Currently supports
        only .txt, .xy, and .csv files.

        Parameters
        ----------
        filename: str
            path to file containing spectral data
        delimiter: str, optional
            character used to delimit columns in file, defaults to whitespace
        xcolumn: int, optional
            column, that should be used as points on x axis,
            defaults to 0 (first column)
        ycolumn: int, optional
            column, that should be used as values on y axis,
            defaults to 1 (second column)

        Returns
        -------
        numpy.array
            two-dimensional numpy array ([[x-values], [y-values]])
            of data type ``float``"""
        # TODO: add type checking of passed file, consider those:
        #     https://github.com/audreyr/binaryornot
        #     https://eli.thegreenplace.net/2011/10/19/\
        #       perls-guess-if-file-is-text-or-binary-implemented-in-python/
        # TODO: add binary files support"""
        self.delimiter = delimiter
        self.xcolumn = xcolumn
        self.ycolumn = ycolumn
        filename = str(filename)
        self.workhorse(filename)  # figure out which method to use
        spc = self.workhorse(filename)  # actual parsing
        return spc

    def initial(self, filename: str):
        super().initial(filename)
        if self.workhorse is self.initial:
            raise ValueError(f"Don't know how to parse file {filename}")

    @ParserBase.state(trigger=r".+\.(?:txt|xy)$")
    def parse_txt(self, file: Path):
        """Loads spectral data from .txt or .xy file to numpy.array.

        Parameters
        ----------
        file: str
            path to file containing spectral data
        delimiter: str, optional
            character used to delimit columns in file, defaults to whitespace
        xcolumn: int, optional
            column, that should be used as points on x axis,
            defaults to 0 (first column)
        ycolumn: int, optional
            column, that should be used as values on y axis,
            defaults to 1 (second column)

        Returns
        -------
        numpy.array
            two-dimensional numpy array ([[x-values], [y-values]])
            of data type 'float'

        Rises
        -----
        ValueError
            if file passed was read to end, but no spectral data was found;
            this includes columns' numbers out of range and usage of
            inappropriate delimiter"""
        with open(file, "r") as txtfile:
            delimiter = self.delimiter
            xcolumn = self.xcolumn
            ycolumn = self.ycolumn
            line = txtfile.readline()
            lineno = 1
            search = True
            while line and search:
                try:
                    values = [v.strip() for v in line.split(delimiter) if v]
                    x, y = float(values[xcolumn]), float(values[ycolumn])
                    search = False
                except (ValueError, TypeError, IndexError) as error:
                    logger.debug(f"Line omitted due to {error}")
                    line = txtfile.readline()
                    lineno += 1
            if not line:
                raise ValueError(
                    f"No spectral data found in file's columns {xcolumn} "
                    f"and {ycolumn}."
                )
            arr = [(x, y)]
            for lineno, line in enumerate(txtfile, start=lineno + 1):
                values = [v.strip() for v in line.split(delimiter) if v]
                arr.append(tuple(map(float, (values[xcolumn], values[ycolumn]))))
        return np.array(list(zip(*arr)))

    @ParserBase.state(trigger=r".+\.csv$")
    def parse_csv(self, file: Path):
        """Loads spectral data from csv file to numpy.array.

        Parameters
        ----------
        file: str
            path to file containing spectral data
        delimiter: str, optional
            character used to delimit columns in file, defaults to ','
        xcolumn: int, optional
            column, that should be used as points on x axis,
            defaults to 0 (first column)
        ycolumn: int, optional
            column, that should be used as values on y axis,
            defaults to 1 (second column)

        Returns
        -------
        numpy.array
            two-dimensional numpy array ([[x-values], [y-values]])
            of data type 'float'"""
        delimiter = self.delimiter
        xcolumn = self.xcolumn
        ycolumn = self.ycolumn
        arr = []
        with open(file, newline="") as csvfile:
            dialect = csv.Sniffer().sniff(csvfile.read(1024), delimiters=delimiter)
            csvfile.seek(0)
            reader = csv.reader(csvfile, dialect)
            for line in reader:
                arr.append(tuple(map(float, (line[xcolumn], line[ycolumn]))))
        return np.array(list(zip(*arr)))

    def parse_spc(self, file):
        """Loads spectral data from spc file to numpy.array.

        Notes
        -----
        This method is not implemented yet, it will raise an error when called.

        Parameters
        ----------
        file: str
            path to file containing spectral data

        Returns
        -------
        numpy.array
            two-dimensional numpy array ([[x-values], [y-values]])
            of data type 'float'

        Raises
        ------
        NotImplementedError
            Whenever called, as this functionality is not implemented yet."""
        # TODO: add support for .spc files
        raise NotImplementedError("Parsing spc files is not implemented yet.")



"""Core functionality of :class:`.DataArray` classes.

This module implements the base class for :class:`.DataArray`\\s and its core
functionality, namely validation of array-like data, along with some helper functions.
To implement a :class:`.DataArray`-like container, subclass the :class:`ArrayBase` class
and use one of the :class:`ArrayProperty` classes to create a validated array-like
instance attribute for your new class. You should also provide *associated_genres* class
attribute to signalize, which genres this new :class:`.DataArray`-like class should be
used for.

The most basic example may look like this:

>>> class MyDataArray(ArrayBase):
...     associated_genres = ("foo",)
...     filenames = ArrayProperty(dtype=str)
...     values = ArrayProperty(check_against="filenames")
...     def __init__(genre, filenames, values, allow_data_inconsistency=False):
...         super().__init__(genre, filenames, values, allow_data_inconsistency)

>>> foo_array = MyDataArray("foo", ["a", "b", "c"], values=[1, 2, 3])

This definition would be almost a re-implementation of what :class:`ArrayBase` already
provides, but is a good starting point for explanation, so lets elaborate on it a
little. :class:`ArrayBase` expects 4 parameters on initialization of its subclass:
*genre* is a genre of data stored, *filenames* is a list of conformer identifiers,
*values* is - not surprisingly - a list of data values for each conformer, and
*allow_data_inconsistency* is a boolean flag that controls process of validation of
array-like attributes.

*filenames* and *values* are :class:`ArrayProperty` instances - values passed to the
constructor as parameters of these names will be checked and validated, and stored as
``numpy.ndarray``\\s. Moreover, filenames will be stored as strings, because we told the
:class:`ArrayProperty` this is our desired data type for this array-like attribute,
using ``dtype=str``. The default data type is ``float``, so *values* will be converted
to floats.

>>> foo_array.filenames
array(["a", "b", "c"], dtype=str)
>>> foo_array.values
array([1.0, 2.0, 3.0], dtype=float)

``check_against="filenames"`` tells :class:`ArrayProperty` to validate *values* using
*filenames* as a reference for desired shape of *values* array. If shape is different
than shape of the reference, :class:`.InconsistentDataError` is raised. If you will deal
with multidimensional data, you can utilize *check_depth* parameter to signalize that
arrays should have identical shapes only to some certain depth, for example
``check_depth=2`` would accept arrays of shapes (10, 20) and (10, 20, 3) but would raise
exception on arrays shaped (10,) and (10, 3). However, in our simple example it wouldn't
make much sense to check more than default depth of 1, since *filenames* have only one
dimension.

>>> MyDataArray("foo", ["a", "b", "c"], values=[1, 2, 3, 4])
Traceback (most recent call last):
     ...
InconsistentDataError: values and filenames must have the same shape up to 1 dimensions.
Arrays of shape (3,) and (4,) were given.

The above exception is also raised if values given to :class:`ArrayProperty` are a
jagged sequence, that is not all entries of the array have identical number of
sub-entries. An example of jagged array would be ``[[1, 2], [3]]``. Data in this format
usually comes from reading calculations of different molecules rather than conformers,
or from corrupted or incomplete output files, so it is not allowed by default. However,
if you are sure that you want to work with such data, you can pass
``allow_data_inconsistency=True`` to your ``MyDataArray`` constructor and
:class:`ArrayProperty` will try to fill-in missing values, producing
``numpy.ma.masked_array`` or at least will ignore inconsistencies. You can chose the
fill value by specifying *fill_value* parameter on :class:`ArrayProperty` instantiation.

Finally we specify ``associated_genres = ("foo",)``, which is the only thing in our
example that's not already defined by :class:`ArrayBase`. This class attribute informs
:class:`.Conformers` object that it should use this :class:`ArrayBase` subclass to
instantiate :class:`.DataArray`-like objects for data genres specified in
``associated_genres``. It must be specified as a tuple of strings, buy may be left
empty, if no genre should be associated with this particular class. However, the main
pourpose of :class:`ArrayBase` is to provide integration with :class:`.Conformers`
machinery - if you wish to use :class:`ArrayProperty`'s validation features only, you
may safely use if in a custom class. It may define ``allow_data_inconsistency``
attribute, but it is optional (``False`` is assumed).

>>> class CustomDataHolder:
...     allow_data_inconsistency=True  # class-level attribute will also work
...     points = ArrayProperty(fill_value=0)
...     def __init__(self, points):
...         self.points = points
...
>>> d = CustomDataHolder(points=((1,2,3),(1,2)))
>>> d.points
masked_array(
  data=[[1.0, 2.0, 3.0],
        [1.0, 2.0, --]],
  mask=[[False, False, False],
        [False, False,  True]],
  fill_value=0)

*genre*, *filenames*, *values*, and *allow_data_inconsistency* are stored on
:class:`ArrayBase` subclass automatically, if ``super().__init__()`` is called. However,
if you introduce any new init parameters, you must bind them to the object by yourself.
Moreover, if you wish to use :class:`.Conformers` automatic initialization of
:class:`ArrayBase` subclasses, you should name those additional parameters with a name
of genre you'd like to be retrived or give them a default value, otherwise
:meth:`.Conformers.arrayed` won't know how to initialize such class.
"""
import inspect
import logging as lgg
from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np

from ..exceptions import InconsistentDataError

# LOGGER
logger = lgg.getLogger(__name__)
logger.setLevel(lgg.DEBUG)


# FUNCTIONS
NestedSequence = Sequence[Union[Any, "NestedSequence"]]


def longest_subsequences(sequences: NestedSequence) -> Tuple[int, ...]:
    """Finds lengths of longest subsequences on each level of given nested sequence.
    Each subsequence should have same number of nesting levels.

    Parameters
    ----------
    sequences : sequence [of sequences [of...]]
        Arbitrarily deep, nested sequence of sequences.

    Returns
    -------
    tuple of ints
        Length of the longest subsequence for each nesting level as a tuple.

    Notes
    -----
    If nesting level in not identical in all subsequences, lengths are reported
    up to first level of non-iterable elements.

    >>> longest_subsequences([[[1, 2]], [[1], 2]])
    (2,)

    Examples
    --------
    >>> longest_subsequences([[[1, 2]], [[1]]])
    (1, 2)
    >>> longest_subsequences([[[1, 2]], [[1], [1], [1]]])
    (3, 2)
    """
    try:
        if any(isinstance(v, str) for v in sequences):
            raise TypeError
        lenghts = [len(v) for v in sequences]
        longest = max(lenghts)
    except (TypeError, ValueError):
        return ()
    try:
        other = longest_subsequences([i for v in sequences for i in v])
    except TypeError:
        return (longest,)  # tuple
    return (longest, *other)


def find_best_shape(jagged: NestedSequence) -> Tuple[int, ...]:
    """Find shape of an array, that could fit arbitrarily deep, jagged, nested sequence
    of sequences. Reported size for each level of nesting is the length of the longest
    subsequence on this level.

    Parameters
    ----------
    jagged : sequence [of sequences [of...]]
        Arbitrarily deep, nested sequence of sequences.

    Returns
    -------
    tuple of ints
        Length of the longest subsequence for each nesting level as a tuple.

    Notes
    -----
    If nesting level in not identical in all subsequences, size is reported
    up to first level of non-iterable elements.

    >>> find_best_shape([[[1, 2]], [[1], 2]])
    (2, 2)

    Examples
    --------
    >>> find_best_shape([[[1, 2]], [[1]]])
    (2, 1, 2)
    >>> find_best_shape([[[1, 2]], [[1], [1], [1]]])
    (2, 3, 2)
    """
    return (len(jagged), *longest_subsequences(jagged))


def flatten(items: NestedSequence, depth: Optional[int] = None) -> Iterator:
    """Yield items from any nested iterable as chain of values up to given *depth*.
    If *depth* is ``None``, yielded sequence is completely flat.

    Parameters
    ----------
    items : NestedSequence
        Arbitrarily deep, nested sequence of sequences.
    depth : int, optional
        How deep should fattening be.

    Yields
    ------
    Any
        Values from *items* as flatted sequence."""
    depth = float("inf") if depth is None else depth
    for x in items:
        should_iter = isinstance(x, Iterable) and not isinstance(x, (str, bytes))
        if should_iter and depth > 1:
            yield from flatten(x, depth - 1)
        else:
            yield x


def _mask(jagged: NestedSequence, shape: tuple) -> np.ndarray:
    """A workhorse of *mask* function, see there.

    Parameters
    ----------
    jagged : sequence [of sequences [of...]]
        Arbitrarily deep, nested sequence of sequences.
    shape : tuple of int
        Shape of an array that would fit *jagged*.

    Returns
    -------
    numpy.array
    Array of booleans, of given shape, indicating if value of same index exist
    in *jagged*.

    Notes
    -----
    Current implementation does not support shapes other than best fitting
    (see :func:`.find_best_shape`).
    """
    if len(shape) == 0:
        mask = np.array([])
    elif len(shape) == 1:
        mask = np.arange(shape[0]) < len(jagged)
    else:
        jagged = [_mask(v, shape[1:]) for v in jagged]
        padding = ((0, shape[0] - len(jagged)),) + ((0, 0),) * (len(shape) - 1)
        mask = np.pad(jagged, padding, "constant", constant_values=False)
    return mask


def mask(jagged: NestedSequence) -> np.ndarray:
    """Returns a numpy.array of booleans, of shape that best fits given jagged nested
    sequence *jagged*. Each boolean value of the output indicates if corresponding value
    exists in *jagged*.

    Parameters
    ----------
    jagged : sequence [of sequences [of...]]
        Arbitrarily deep, nested sequence of sequences.

    Returns
    -------
    numpy.array of bool
        Array of booleans, of shape that best fits *jagged*, indicating if value of
        same index exist in *jagged*.

    Notes
    -----
    To use output as a mask of numpy.ma.masked_array, it should be inverted.
    >>> np.ma.array(values, mask=~mask(jagged))

    Examples
    --------
    >>> mask([[1, 2], [1]])
    array([[True, True], [True, False]])
    >>> mask([[1, 2], []])
    array([[True, True], [False, False]])
    >>> mask([[[1], []], [[2, 3]]])
    array([[[True, False], [False, False]], [[True, True], [False, False]]])

    """
    return _mask(jagged, find_best_shape(jagged))


def to_masked(
    jagged: NestedSequence,
    dtype: Optional[type] = None,
    fill_value: Optional[Any] = None,
) -> np.ma.core.MaskedArray:
    """Convert jagged, arbitrarily deep, nested sequence to numpy.ma.masked_array
    with missing entries masked.

    Parameters
    ----------
    jagged : sequence [of sequences [of...]]
        Arbitrarily deep, nested sequence of sequences.
    dtype : type, optional
        Data type of the output. If *dtype* is ``None``, the type of the data is figured
        out by numpy machinery.
    fill_value : scalar, optional
        Value used to fill in the masked values when necessary. If ``None``, a default
        based on the data-type is used.

    Returns
    -------
    numpy.ma.core.MaskedArray
        Given *jagged* converted to numpy.ma.masked_array with missing entries masked.

    Raises
    ------
    ValueError
        If jagged sequence has inconsistent number of dimensions.

    Examples
    --------
    >>> to_masked([[1, 2], [1]])
    array(data=[[1, 2], [1, --]], mask=[[True, True], [True, False]])
    >>> to_masked([1, [1]])
    Traceback (most recent call last):
    ValueError: Cannot convert to masked array: jagged sequence has inconsistent
    number of dimensions.
    """
    if not jagged:
        return np.ma.array([])
    array, shape, done, lengths = jagged, (len(jagged),), False, []
    while not done:
        # iterate over each level of nesting, accumulating partially flatted array
        each_is_vector = all(
            isinstance(x, Iterable) and not isinstance(x, (str, bytes)) for x in array
        )
        each_is_scalar = all(
            not isinstance(x, Iterable) or isinstance(x, (str, bytes)) for x in array
        )
        if each_is_vector:
            # if each entry is vector, find these vectors' length
            lengths = [len(x) for x in array]
            # longest vector defines current dimension's shape
            shape = shape + (max(lengths),)
            # flatten one level
            array = [v for x in array for v in x]
        elif each_is_scalar:
            # if each entry is scalar, we're done
            done = True
        else:
            # if scalar and vector entries mixed, we raise an error
            raise ValueError(
                "Cannot convert to masked array: jagged sequence has inconsistent "
                "number of dimensions."
            )
    mask = _mask(jagged, shape)
    output = np.ma.zeros(shape, dtype=dtype)
    # fill output with values
    output[mask] = array
    # mask for numpy.ma.masked_array must be inverted (*True* masks value)
    return np.ma.array(output, mask=~mask, dtype=dtype, fill_value=fill_value)


# CLASSES
class ArrayProperty(property):
    """Property, that validates array-like value given to its setter and stores it as
    ``numpy.ndarray``.

    Value given to property setter is:

    1. (optionally) sanitized with user-provided sanitizer function;
    2. (optionally) compared with another array-like attribute of the owner
       regarding their shape;
    3. transformed to ``numpy.ndarray`` of desired data type;
    4. stored in owner's ``__dict__``.

    Setting, getting and deletition of the value may be customized using standard
    :attr:`~.ArrayProperty.setter`, :attr:`~.ArrayProperty.getter` and
    :attr:`~.ArrayProperty.deleter` decorators. Additionally, :class:`ArrayProperty`
    provides an :attr:`.ArrayProperty.sanitizer` decorator. If sanitizer function is
    provided, it is called as a first step of data validation and should return
    sanitized array-like value (given original value as a positional parameter).

    Validation regarding shape of the value is triggered if *check_against*
    parameter is provided. It should be a name of owner's other array-like attribute
    as a string. Shape of the value is than compared to the shape of this reference
    attribute. If shapes are not identical up to the first *check_depth* dimensions,
    :class:`.InconsistentDataError` is raised.

    Value is always transformed to ``numpy.ndarray`` of specified *dtype* (``float`` by
    default.) If such conversion cannot be done because value is a jagged array,
    :class:`.InconsistentDataError` will be raised. However, if owner allows for data
    inconsistency by defining ``owner.allow_data_inconsistency = True``, non-matching
    shapes will be ignored and jugged arrays will be padded with *fill_value* and stored
    as ``numpy.ma.masked_array``.
    """

    def __init__(
        self,
        fget: Optional[Callable[[Any], np.ndarray]] = None,
        fset: Optional[Callable[[Any, Sequence], None]] = None,
        fdel: Optional[Callable[[Any], None]] = None,
        doc: str = None,
        dtype: type = float,
        check_against: Optional[str] = None,
        check_depth: int = 1,
        fill_value: Any = 0,
        fsan: Optional[Callable[[Sequence], Sequence]] = None,
    ):
        """
        Parameters
        ----------
        fget
            Custom getter for attribute. Default one just returns the stored value.
        fset
            Custom setter for attribute. Default one stores validated values in
            instance's ``__dict__``.
        fdel
            Custom deleter for attribute. Deleting attribute is not supported by
            default.
        doc
            Attribute's docstring.
        dtype
            Data type of elements of this array-like attribute.
        check_against
            Which other instance's attribute should be used as a reference for array's
            shape. If shape of this attribute and reference attribute's are different,
            an exception is raised. Only first *check_depth* dimensions are compared.
        check_depth
            How many dimensions should be compared when checking shape of the array.
        fill_value
            If values are a jagged array and ``instance.allow_data_inconsistency is
            True``, this value will be passed to ``numpy.ma.masked_array`` constructor
            as a *fill_value*.
        fsan
            Custom sanitizer for attribute. "Sanitizer" is here understood as a function
            that transforms value received by the setter, before the value is validated
            (checked for corectness) and stored on the instance. *fsan* should return a
            sanitized value.
        """
        super().__init__(fget=fget, fset=fset, fdel=fdel, doc=doc)
        self.dtype = dtype
        self.check_against = check_against
        self.check_depth = check_depth
        self.fill_value = fill_value
        self.fsan = fsan

    def __set_name__(self, objtype, name):
        self.name = name

    def __set__(self, instance, values):
        if self.fsan is not None:
            values = self.fsan(values)
        values = self.check_input(instance, values)
        if self.fset is not None:
            super().__set__(instance, values)
        else:
            vars(instance)[self.name] = values

    def __get__(self, instance, objtype=None):
        if instance is None:
            return self
        if self.fget is not None:
            return super().__get__(instance)
        else:
            return vars(instance)[self.name]

    def __call__(self, fget):
        new = self.getter(fget)
        if self.__doc__ is None:
            new.__doc__ = fget.__doc__
        return new

    def getter(self, fget: Optional[Callable[[Any], Sequence]]):
        """Descriptor to change the getter on an :class:`ArrayProperty`."""
        return type(self)(
            fget,
            self.fset,
            self.fdel,
            self.__doc__,
            self.dtype,
            self.check_against,
            self.fill_value,
            self.fsan,
        )

    def setter(self, fset: Optional[Callable[[Any, Sequence], None]]):
        """Descriptor to change the setter on an :class:`ArrayProperty`."""
        return type(self)(
            self.fget,
            fset,
            self.fdel,
            self.__doc__,
            self.dtype,
            self.check_against,
            self.fill_value,
            self.fsan,
        )

    def deleter(self, fdel: Optional[Callable[[Any], None]]):
        """Descriptor to change the deleter on an :class:`ArrayProperty`."""
        return type(self)(
            self.fget,
            self.fset,
            fdel,
            self.__doc__,
            self.dtype,
            self.check_against,
            self.fill_value,
            self.fsan,
        )

    def sanitizer(self, fsan: Optional[Callable[[Sequence], Sequence]]):
        """Descriptor to change the sanitizer on an :class:`ArrayProperty`. Function
        given as parameter should take one positional argument and return sanitized
        values. If any sanitizer is provided, it is always called with *values* given to
        :class:`ArrayProperty` setter. Sanitation is performed before `.check_input()`
        is called."""
        return type(self)(
            self.fget,
            self.fset,
            self.fdel,
            self.__doc__,
            self.dtype,
            self.check_against,
            self.fill_value,
            fsan,
        )

    def check_shape(self, instance: Any, values: Sequence):
        """Raises an error if *values* have different shape than attribute specified
        as :attr:`.check_against`.
        """
        allow = getattr(instance, "allow_data_inconsistency", False)
        if self.check_against:
            attr_value = getattr(instance, self.check_against)
            try:
                ref_shape = attr_value.shape
            except AttributeError:
                ref_shape = find_best_shape(attr_value)
            best_shape = find_best_shape(values)
            shapes_same = (
                ref_shape[: self.check_depth] == best_shape[: self.check_depth]
            )
            if not shapes_same and not allow:
                raise ValueError(
                    f"{self.name} and {self.check_against} must have the same shape "
                    f"up to {self.check_depth} dimensions. Arrays of shape "
                    f"{best_shape} and {ref_shape} were given."
                )

    def check_input(self, instance: Any, values: Sequence) -> np.ndarray:
        """Checks if *values* given to setter have same length as attribute specified
        with :attr:`.check_against`.

        Parameters
        ----------
        instance
            Instance of owner class.
        values
            Values to validate.

        Returns
        -------
        numpy.ndarray
            Validated values.

        Raises
        ------
        ValueError
            If :attr:`.check_against` is not None and list of given values have
            different length than getattr(*instance*, :attr:`.check_against`). If given
            list of values cannot be converted to :attr:`.dtype` type.
        InconsistentDataError
            If *values* is list of lists of varying size and instance doesn't allow
            data inconsistency.
        """
        self.check_shape(instance, values)
        allow = getattr(instance, "allow_data_inconsistency", False)
        try:
            return np.array(values, dtype=self.dtype)
        except ValueError as error:
            genre = getattr(instance, "genre", "<unknown>")
            if "convert" in error.args[0]:
                raise
            elif not allow:
                error_msg = (
                    f"{instance.__class__.__name__} of genre "
                    f"{genre} with unequal number of values "
                    f"for conformer requested."
                )
                raise InconsistentDataError(error_msg) from error
            else:
                values = to_masked(values, dtype=self.dtype, fill_value=self.fill_value)
                logger.info(
                    f"{genre} values' lists were appended with zeros to "
                    f"match length of longest entry."
                )
                return values


class JaggedArrayProperty(ArrayProperty):
    """:class:`ArrayProperty` for storing intentionally jagged arrays of data.
    :class:`.InconsistentDataError` is only raised if :meth:`ArrayProperty.check_shape`
    fails. Given values are converted to masked array and expanded as needed, regardless
    value of *allow_data_inconsistency* attribute."""

    def check_input(self, instance: Any, values: Sequence) -> np.ndarray:
        self.check_shape(instance, values)
        return to_masked(values, dtype=self.dtype, fill_value=self.fill_value)


class CollapsibleArrayProperty(ArrayProperty):
    """:class:`ArrayProperty` that stores only one value, if all entries are
    identical."""

    def __init__(
        self,
        fget: Optional[Callable[[Any], np.ndarray]] = None,
        fset: Optional[Callable[[Any, Sequence], None]] = None,
        fdel: Optional[Callable[[Any], None]] = None,
        doc: str = None,
        dtype: type = float,
        check_against: Optional[str] = None,
        check_depth: int = 1,
        fill_value: Any = 0,
        fsan: Optional[Callable[[Sequence], Sequence]] = None,
        strict: bool = False,
    ):
        """
        Parameters
        ----------
        fget
            Custom getter for attribute. Default one just returns the stored value.
        fset
            Custom setter for attribute. Default one stores validated values in
            instance's ``__dict__``.
        fdel
            Custom deleter for attribute. Deleting attribute is not supported by
            default.
        doc
            Attribute's docstring.
        dtype
            Data type of elements of this array-like attribute.
        check_against
            Which other instance's attribute should be used as a reference for array's
            shape. If shape of this attribute and reference attribute's are different,
            an exception is raised. Only first *check_depth* dimensions are compared.
        check_depth
            How many dimensions should be compared when checking shape of the array.
        fill_value
            If values are a jagged array and ``instance.allow_data_inconsistency is
            True``, this value will be passed to ``numpy.ma.masked_array`` constructor
            as a *fill_value*.
        fsan
            Custom sanitizer for attribute. "Sanitizer" is here understood as a function
            that transforms value received by the setter, before the value is validated
            (checked for corectness) and stored on the instance. *fsan* should return a
            sanitized value.
        strict
            Boolean flag indicating if :meth:`.check_input` should disallow values that
            are not all identical. If *strict* is ``True`` it will raise
            :class:`.InconsistentDataError` when setter is given such values. Defaults
            to ``False``.
        """
        self.strict = strict
        super().__init__(
            fget=fget,
            fset=fset,
            fdel=fdel,
            doc=doc,
            dtype=dtype,
            check_against=check_against,
            check_depth=check_depth,
            fill_value=fill_value,
            fsan=fsan,
        )

    def check_shape(self, instance: Any, values: Sequence):
        """Raises an error if *values* have different shape than attribute specified as
        :attr:`.check_against`. Accepts values with size of first dimension equal to 1,
        even if it is not identical to the size of the first dimension of said
        attribute.
        """
        allow = getattr(instance, "allow_data_inconsistency", False)
        if self.check_against:
            attr_value = getattr(instance, self.check_against)
            try:
                ref_shape = attr_value.shape
            except AttributeError:
                ref_shape = find_best_shape(attr_value)
            best_shape = find_best_shape(values)
            first_important = int(best_shape[0] == 1)
            shapes_same = (
                ref_shape[first_important : self.check_depth]
                == best_shape[first_important : self.check_depth]
            )
            if not shapes_same and not allow:
                raise ValueError(
                    f"{self.name} and {self.check_against} must have the same shape "
                    f"up to {self.check_depth} dimensions. Arrays of shape "
                    f"{best_shape} and {ref_shape} were given."
                )

    def check_input(self, instance: Any, values: Union[Sequence, Any]) -> np.ndarray:
        """If given *values* is not iterable or is of type ``str`` it is returned
        without change. Otherwise it is validated using
        :meth:`ArrayProperty.check_input`, and collapsed to single value if all values
        are identical. If values are non-uniform and instance doesn't allow data
        inconsistency, :class:`.InconsistentDataError` is raised.

        Parameters
        ----------
        instance
            Instance of owner class.
        values
            Values to validate.

        Returns
        -------
        numpy.ndarray or any
            Validated array or single value.

        Raises
        ------
        ValueError
            If :attr:`ArrayProperty.check_against` is not None and list of given values
            have different length than getattr(*instance*,
            :attr:`ArrayProperty.check_against`). If given list of values cannot be
            converted to :attr:`ArrayProperty.dtype` type.
        InconsistentDataError
            If *values* is list of lists of varying size and instance doesn't allow
            data inconsistency.
            If property is declared as strict, given *values* are non-uniform
            and instance doesn't allow data inconsistency.
        """
        if not isinstance(values, Iterable) or isinstance(values, str):
            return np.array([values], dtype=self.dtype)
        values = super().check_input(instance, values)
        allow = getattr(instance, "allow_data_inconsistency", False)
        try:
            all_same = (values == values[0]).all()
        except IndexError:
            return np.array([], dtype=self.dtype)
        if all_same:
            # return only first element, but keep dimensionality
            return values[np.newaxis, 0]
        elif self.strict and not allow:
            raise InconsistentDataError(
                "List of non-uniform values given to CollapsibleArrayProperty setter."
            )
        else:
            return values


_ARRAY_CONSTRUCTORS = {}
"""Registry of classes used to represent certain data genres.
It is a dictionary of {str: ArrayBase}."""


class ArrayBase(ABC):
    """Base class for data holding objects.

    It provides an automatic registration of its subclasses as a
    :class:`.DataArray`-like representations of all :attr:`.associated_genres` declared
    by said subclass. A subclass should provide an :attr:`.associated_genres` class
    attribute, even if it's not supposed to be directly instantiated with data for any
    genre, it should be an empty tuple in such case. Otherwise,
    :attr:`.associated_genres` should be a tuple of genre names as strings.

    This base class provides the most basic set of attributes, a
    :class:`.DataArray`-like object should implement, listed in the Parameters section.
    """

    @property
    @classmethod
    @abstractmethod
    def associated_genres(cls) -> Tuple[str, ...]:
        return tuple()

    associated_genres.__doc__ = """Genres associated with subclassing class.

    Should be provided by subclass as class-level attribute. It will be used to
    determine what class to use to represent data of particular genre when requested
    *via* :meth:`.Conforemrs.arrayed` method. May be an empty sequence, if subclass
    is not intended to be used directly by ``tesliper``'s machinery.
    """

    def __init_subclass__(cls, **kwargs):
        if isinstance(cls.associated_genres, str):
            raise TypeError(
                f"`{cls.__name__}.associated_genres` should be a sequence of str, "
                "not just str."
            )
        global _ARRAY_CONSTRUCTORS
        if cls.associated_genres is ArrayBase.associated_genres:
            raise TypeError(
                f"`{cls.__name__}` must provide `associated_genres` class attribute. "
                "It may be an empty tuple if no genres should be associated with "
                "this class."
            )
        _ARRAY_CONSTRUCTORS.update((genre, cls) for genre in cls.associated_genres)

    def __init__(
        self,
        genre: str,
        filenames: Sequence[str],
        values: Sequence,
        allow_data_inconsistency: bool = False,
    ):
        """
        Parameters
        ----------
        genre
            Name of the data genre that *values* represent.
        filenames
            Sequence of conformers' identifiers.
        values
            Sequence of values for *genre* for each conformer in *filenames*.
        allow_data_inconsistency
            Flag signalizing if instance should allow data inconsistency (see
            :class:`.ArrayProperty` for details).
        """
        self.genre = genre
        self.allow_data_inconsistency = allow_data_inconsistency
        self.filenames = filenames
        self.values = values

    filenames = ArrayProperty(dtype=str)
    values = ArrayProperty(check_against="filenames")

    def get_repr_args(self) -> Dict[str, Any]:
        """Returns dictionary that can be used as keword-value pairs to instantiate
        identical object."""
        signature = inspect.signature(type(self))
        args = {
            name: getattr(self, name)
            if hasattr(self, name)
            else (param.default if param.default is not signature.empty else None)
            for name, param in signature.parameters.items()
        }
        return args

    @classmethod
    def get_init_params(cls) -> Dict[str, Union[str, inspect.Parameter]]:
        """Returns parameters used to instantiate this class. *genre* is a genre of data
        array that is to be instantiated."""
        signature = inspect.signature(cls)
        parameters = signature.parameters.copy()
        return parameters

    def __repr__(self):
        args = [
            f"{name}={repr(arg) if isinstance(arg, str) else arg}"
            for name, arg in self.get_repr_args().items()
        ]
        return f"{type(self).__name__}({', '.join(args)})"

    def __str__(self):
        return (
            f"[{type(self).__name__} of genre '{self.genre}', "
            f"{self.filenames.size} conformers]"
        )

    def __len__(self):
        return len(self.filenames)

    def __bool__(self):
        return self.filenames.size != 0


class DependentParameter(inspect.Parameter):
    """A parameter that depends on the genre of data array. It provies a
    :attr:`._genre_getter` callable attribute that is used to provide a name of data
    genre that should be used for this parameter.

    It is hashable as the original :class:`inspect.Parameter`, however it must be
    remembered that Python hashes functions based on their identity."""

    __slots__ = ("_name", "_kind", "_default", "_annotation", "_genre_getter")

    def __init__(
        self,
        name: str,
        kind: inspect._ParameterKind,
        genre_getter: Callable[[str], str],
        *,
        default: Any = inspect._empty,
        annotation: Any = inspect._empty,
    ) -> None:
        super().__init__(name, kind, default=default, annotation=annotation)
        self._genre_getter = genre_getter

    @property
    def genre_getter(self):
        """Should be a function that given a genre of data array being instantiated,
        returns a genre that should be used for this parameter."""
        return self._genre_getter

    def __reduce__(self):
        return (
            type(self),
            (self._name, self._kind, self._genre_getter),
            {
                "_default": self._default,
                "_annotation": self._annotation,
            },
        )

    @classmethod
    def from_parameter(
        cls, parameter: inspect.Parameter, genre_getter: Callable[[str], str]
    ):
        """Casts given :class:`inspect.Parameter` instance to this class."""
        return cls(
            name=parameter.name,
            kind=parameter.kind,
            genre_getter=genre_getter,
            annotation=parameter.annotation,
            default=parameter.default,
        )

    def replace(
        self,
        *,
        name=inspect._void,
        kind=inspect._void,
        genre_getter=inspect._void,
        annotation=inspect._void,
        default=inspect._void,
    ):
        """Creates a customized copy of the :class:`.DependentParameter`."""

        if name is inspect._void:
            name = self._name

        if kind is inspect._void:
            kind = self._kind

        if annotation is inspect._void:
            annotation = self._annotation

        if default is inspect._void:
            default = self._default

        if genre_getter is inspect._void:
            genre_getter = self._genre_getter

        return type(self)(
            name,
            kind,
            genre_getter,
            default=default,
            annotation=annotation,
        )

    def __hash__(self):
        return hash(
            (self.name, self.kind, self.genre_getter, self.annotation, self.default)
        )

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, DependentParameter):
            return NotImplemented
        return (
            self._name == other._name
            and self._kind == other._kind
            and self._default == other._default
            and self._annotation == other._annotation
            and self._genre_getter == other._genre_getter
        )



import logging as lgg
import tkinter.ttk as ttk

# LOGGER
logger = lgg.getLogger(__name__)


# CLASSES
class CollapsiblePane(ttk.LabelFrame):
    """LabelFrame that can be collapsed to show only its label.
    Clicking it will toggle between collapsed and unraveled states.

    Adds *collapsed* parameter to `.configure()` setting it to *True* will hide
    *CollapsiblePane*'s content, and setting to *False* will do the opposite.
    """

    def __init__(
        self,
        parent,
        text="",
        collapsed=False,
        collapsed_mark="▷",
        unraveled_mark="▽",
        **kwargs,
    ):
        super().__init__(parent, **kwargs)

        self.collapsed = collapsed
        self.text = text
        self.collapsed_mark = collapsed_mark
        self.unraveled_mark = unraveled_mark

        ttk.Frame(self).grid(column=0, row=0, sticky="nwe")
        self.content = ttk.Frame(self)
        self.content.grid(column=0, row=0, sticky="news")
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        self.bind("<Button-1>", self.toggle)
        self.configure(collapsed=self.collapsed)

    @property
    def mark(self):
        return self.collapsed_mark if self.collapsed else self.unraveled_mark

    def configure(self, cnf=None, **kwargs):
        if "text" in kwargs:
            self.text = kwargs["text"]
        if "collapsed" in kwargs:
            self.collapsed = kwargs.pop("collapsed")
            kwargs["text"] = f"{self.mark} {self.text}"
            self.content.grid_remove() if self.collapsed else self.content.grid()
        content_config = {
            cfg: kwargs[cfg]
            for cfg in (
                "background",
                "cursor",
                "highlightbackground",
            )
            if cfg in kwargs
        }
        self.content.configure(**content_config)
        super().configure(cnf, **kwargs)

    def toggle(self, _event=None):
        if _event is not None:
            logger.debug(f"Event caught by {self}.toggle handler.")
        self.configure(collapsed=not self.collapsed)



import pytest

from tesliper import glassware as gw
from tesliper.exceptions import TesliperError
from tesliper.glassware import conformers as cf


@pytest.fixture
def empty():
    return cf.Conformers()


@pytest.fixture
def single():
    return cf.Conformers(bla={"data": [1, 2, 3, 4]})


@pytest.fixture
def base():
    return {
        "normal_termination": True,
        "version": "Gaussian 09, Revision E.01",
        "command": "# opt freq=vcd B3LYP/Def2TZVP",
        "optimization_completed": True,
        "charge": 0,
        "multiplicity": 1,
        "input_geom": [],
        "stoichiometry": "CH3F",
        "last_read_atoms": (6, 1, 1, 1, 9),
        "last_read_geom": [],
        "freq": [1, 2, 3, 4, 5],
        "mass": [2, 1, 3, 2, 1],
        "iri": [27, 8, 569, 1, 3],
        "emang": [1.9, 1.4, 4.4, 1.1, 1.3],
        "wavelen": [231, 144, 137],
        "vdip": [0.0003, 0.008, 0.0022],
        "transitions": [
            ((20, 25, 0.1),),
            ((21, 25, -0.2), (23, 25, 0.6)),
            ((13, 25, -0.1), (15, 25, 0.3), (18, 25, -0.4)),
        ],
        "zpecorr": 0.5,
        "zpe": -200,
        "scf": 0,
    }


@pytest.fixture
def full(base):
    incom = {**base, "zpe": -260}
    del incom["scf"]
    return cf.Conformers(
        base=base,
        noopt={**base, "zpe": -210, "optimization_completed": False},
        imag={**base, "zpe": -220, "freq": [-1, 2, 3, 4, 5]},
        stoich={**base, "zpe": -230, "stoichiometry": "CH3F_other"},
        term={**base, "zpe": -240, "normal_termination": False},
        size={**base, "zpe": -250, "mass": [1, 2, 3]},
        incom=incom,
    )


@pytest.fixture
def jagged(full):
    full["size"]["excess"] = 1
    return full


def test_instantiation(empty, single):
    assert empty.kept == []
    assert empty.filenames == []
    assert single.kept == [True]
    assert single.filenames == ["bla"]
    assert isinstance(single["bla"], dict)


def test_setitem_with_dict(single):
    single["foo"] = {"data": [1, 2, 3, 4]}
    assert len(single) == 2
    assert single.kept == [True, True]
    assert single.filenames == ["bla", "foo"]


def test_setitem_with_zip(single):
    single["foo"] = zip(("data",), ([1, 2, 3, 4],))
    assert len(single) == 2
    assert single.kept == [True, True]
    assert single.filenames == ["bla", "foo"]


def test_setitem_with_tuple(single):
    single["foo"] = (("data", [1, 2, 3, 4]),)
    assert len(single) == 2
    assert single.kept == [True, True]
    assert single.filenames == ["bla", "foo"]


def test_setitem_with_invalid_type(single):
    with pytest.raises(TypeError):
        single["ham"] = (1, 2)
    with pytest.raises(TypeError):
        single["ham"] = 1


def test_setitem_with_invalid_value(single):
    with pytest.raises(ValueError):
        single["ham"] = ((1, 2, 3),)
    with pytest.raises(ValueError):
        single["ham"] = ((1,),)


def test_delitem_single(single):
    del single["bla"]
    assert single.kept == []
    assert single.filenames == []


def test_delitem_many(full):
    del full["imag"]
    assert full.kept == [True] * 6
    assert full.filenames == ["base", "noopt", "stoich", "term", "size", "incom"]
    assert [full._indices[k] for k in full.keys()] == list(range(6))


def test_update_with_kwargs(single):
    single.update(bla2={"data": [1, 2, 3, 4]})
    assert len(single) == 2
    assert single.kept == [True, True]
    assert single.filenames == ["bla", "bla2"]
    assert single._indices["bla2"] == 1
    assert single["bla2"] == {"data": [1, 2, 3, 4]}


def test_update_with_dict(single):
    single.update({"bla2": {"data": [1, 2, 3, 4]}})
    assert len(single) == 2
    assert single.kept == [True, True]
    assert single.filenames == ["bla", "bla2"]
    assert single._indices["bla2"] == 1
    assert single["bla2"] == {"data": [1, 2, 3, 4]}


def test_update_with_zip(single):
    single.update(zip(("bla2",), ({"data": [1, 2, 3, 4]},)))
    assert len(single) == 2
    assert single.kept == [True, True]
    assert single.filenames == ["bla", "bla2"]
    assert single._indices["bla2"] == 1
    assert single["bla2"] == {"data": [1, 2, 3, 4]}


def test_update_with_tuple(single):
    single.update((("bla2", {"data": [1, 2, 3, 4]}),))
    assert len(single) == 2
    assert single.kept == [True, True]
    assert single.filenames == ["bla", "bla2"]
    assert single._indices["bla2"] == 1
    assert single["bla2"] == {"data": [1, 2, 3, 4]}


def test_update_new(single):
    single.update(bla2={"data": [5, 6, 7, 8]})
    assert len(single) == 2
    assert single.kept == [True, True]
    assert single.filenames == ["bla", "bla2"]
    assert single._indices["bla2"] == 1
    assert single["bla2"] == {"data": [5, 6, 7, 8]}


def test_update_existing_with_new_data(single):
    single.update(bla={"data_new": "new"})
    assert len(single) == 1
    assert single.kept == [True]
    assert single.filenames == ["bla"]
    assert single._indices == {"bla": 0}
    assert single["bla"] == {"data": [1, 2, 3, 4], "data_new": "new"}


def test_update_existing_and_new(single):
    single.update(bla={"data_new": "new"}, bla2={"data": [1, 2, 3, 4]})
    assert len(single) == 2
    assert single.kept == [True, True]
    assert single.filenames == ["bla", "bla2"]
    assert single._indices == {"bla2": 1, "bla": 0}
    assert single["bla2"] == {"data": [1, 2, 3, 4]}
    assert single["bla"] == {"data": [1, 2, 3, 4], "data_new": "new"}


def test_update_repeated(single):
    single.update({"bla": {"data": "new"}}, bla={"other": "foo"})
    assert len(single) == 1
    assert single.kept == [True]
    assert single.filenames == ["bla"]
    assert single["bla"]["data"] == "new"
    assert single["bla"]["other"] == "foo"


def test_update_banned(single):
    with pytest.raises(TypeError):
        single.update(foo=7)
    with pytest.raises(TypeError):
        single.update({"foo": 7})


def test_pop_single(single):
    _ = single.pop("bla")
    assert not single
    assert not single.filenames
    assert not single._indices
    assert not single.kept


@pytest.mark.parametrize(
    "key", ["base", "noopt", "imag", "stoich", "term", "size", "incom"]
)
def test_pop(full, key):
    full.kept = [True, False, True, False, True, False, True]
    expected_filenames, expected_kept = zip(
        *[(n, k) for n, k in zip(full.filenames, full.kept) if n is not key]
    )
    _ = full.pop(key)
    assert key not in full
    assert full.filenames == list(expected_filenames)
    assert full.kept == list(expected_kept)
    assert full._indices == {n: i for i, n in enumerate(expected_filenames)}


@pytest.mark.parametrize("last", [True, False])
def test_popitem_single(single, last):
    pair = single.popitem(last=last)
    assert pair[0] == "bla"
    assert not single
    assert not single.filenames
    assert not single._indices
    assert not single.kept


@pytest.mark.parametrize("last", [True, False])
def test_popitem(full, last):
    full.kept = [True, False, True, False, True, False, True]
    expected_filenames = full.filenames[:-1] if last else full.filenames[1:]
    expected_kept = full.kept[:-1] if last else full.kept[1:]
    expected_key = full.filenames[-1] if last else full.filenames[0]
    pair = full.popitem(last=last)
    assert pair[0] == expected_key
    assert expected_key not in full
    assert full.filenames == list(expected_filenames)
    assert full.kept == list(expected_kept)
    assert full._indices == {n: i for i, n in enumerate(expected_filenames)}


@pytest.mark.parametrize("last", [True, False])
@pytest.mark.parametrize(
    "key", ["base", "noopt", "imag", "stoich", "term", "size", "incom"]
)
def test_move_to_end(full, key, last):
    full.move_to_end(key, last=last)
    assert full.filenames[0 if not last else -1] == key
    assert all(k == n for k, n in zip(full, full.filenames))
    assert all(full._indices[k] == i for i, k in enumerate(full.filenames))


def test_arrayed_default_parameter(full):
    zpe = full.arrayed("zpe")
    assert 298.15 == zpe.t


def test_arrayed_override_default_parameter(full):
    zpe = full.arrayed("zpe", t=300)
    assert 300 == zpe.t


def test_arrayed_unknown_kwarg_raises(full):
    with pytest.raises(TypeError):
        full.arrayed("zpe", unknown=1)


def test_arrayed_not_strict_ignores(full):
    # only checking if doesn't raise an exception
    _ = full.arrayed("zpe", strict=False, unknown=1)


def test_arrayd_empty(full):
    gib = full.arrayed("gib")
    assert [] == gib.filenames.tolist()
    assert [] == gib.values.tolist()
    assert 298.15 == gib.t


def test_arrayed_empty_spectral(full):
    arr = full.arrayed("ramact")
    assert [] == arr.filenames.tolist()
    assert [] == arr.values.tolist()
    assert [] == arr.frequencies.tolist()
    assert 298.15 == arr.t


def test_arrayed_types(full):
    zpe = full.arrayed("zpe")
    assert gw.Energies is type(zpe)
    normal_termination = full.arrayed("normal_termination")
    assert gw.BooleanArray is type(normal_termination)
    command = full.arrayed("command")
    assert gw.InfoArray is type(command)
    iri = full.arrayed("iri")
    assert gw.VibrationalActivities is type(iri)
    ramact = full.arrayed("ramact")
    assert gw.ScatteringActivities is type(ramact)
    vdip = full.arrayed("vdip")
    assert gw.ElectronicActivities is type(vdip)
    emang = full.arrayed("emang")
    assert gw.VibrationalData is type(emang)
    charge = full.arrayed("charge")
    assert gw.IntegerArray is type(charge)
    filenames = full.arrayed("filenames")
    assert gw.FilenamesArray is type(filenames)


def test_arrayed_trimmed(full):
    arr = full.arrayed("zpe")
    assert arr.values.shape == (len(full),)
    full.kept[2] = False
    arr = full.arrayed("zpe")
    assert arr.values.shape == (len(full) - 1,)
    assert list(arr.values) == [-200, -210, -230, -240, -250, -260]


def test_arrayed_unknown(full):
    with pytest.raises(ValueError):
        full.arrayed("nohave")


def test_arrayed_no_data(full):
    del full["incom"]["wavelen"]
    with pytest.raises(TesliperError):
        full.arrayed("vdip")


def test_trim_not_optimized(full):
    full.trim_not_optimized()
    assert [True, False, True, True, True, True, True] == full.kept


def test_trim_not_optimized_previously_trimmed(full):
    full.kept = [False, True, True, True, True, True, True]
    full.trim_not_optimized()
    assert [False, False, True, True, True, True, True] == full.kept


def test_trim_imaginary(full):
    full.trim_imaginary_frequencies()
    assert [True, True, False, True, True, True, True] == full.kept


def test_trim_imaginary_previously_trimmed(full):
    full.kept = [False, True, True, True, True, True, True]
    full.trim_imaginary_frequencies()
    assert [False, True, False, True, True, True, True] == full.kept


def test_trim_stoichiometry(full):
    full.trim_non_matching_stoichiometry()
    assert [True, True, True, False, True, True, True] == full.kept


def test_trim_stoichiometry_previously_trimmed(full):
    full.kept = [False, True, True, True, True, True, True]
    full.trim_non_matching_stoichiometry()
    assert [False, True, True, False, True, True, True] == full.kept


def test_trim_termination(full):
    full.trim_non_normal_termination()
    assert [True, True, True, True, False, True, True] == full.kept


def test_trim_termination_previously_trimmed(full):
    full.kept = [False, True, True, True, True, True, True]
    full.trim_non_normal_termination()
    assert [False, True, True, True, False, True, True] == full.kept


def test_trim_termination_missing(full):
    del full["base"]["normal_termination"]
    full.trim_non_normal_termination()
    assert [False, True, True, True, False, True, True] == full.kept


def test_trim_inconsistent_sizes(full):
    full.trim_inconsistent_sizes()
    assert [True, True, True, True, True, False, True] == full.kept


def test_trim_inconsistent_sizes_previously_trimmed(full):
    full.kept = [False, True, True, True, True, True, True]
    full.trim_inconsistent_sizes()
    assert [False, True, True, True, True, False, True] == full.kept


def test_trim_incomplete_default_previously_trimmed(full):
    full.kept = [False, True, True, True, True, True, True]
    full.trim_incomplete()
    assert [False, True, True, True, True, True, False] == full.kept


def test_trim_incomplete_default(full):
    full.trim_incomplete()
    assert [True, True, True, True, True, True, False] == full.kept
    full.kept = [True, True, True, True, True, True, True]
    del full["base"]["zpe"]
    full.trim_incomplete()
    assert [False, True, True, True, True, True, False] == full.kept


def test_trim_incomplete_wanted(full):
    del full["base"]["zpe"]
    full.trim_incomplete(wanted=["zpe"])
    assert [False, True, True, True, True, True, True] == full.kept
    full.kept = [True, True, True, True, True, True, True]
    full.trim_incomplete(wanted=["scf"])
    assert [True, True, True, True, True, True, False] == full.kept


def test_trim_incomplete_strict(full):
    m = gw.Conformers(
        one={"a": 1, "b": 2},
        two={"a": 1, "c": 3},
        three={"a": 1, "d": 3},
    )
    assert [True, True, True] == m.kept
    m.trim_incomplete(wanted=["a", "b", "c"])
    assert [True, False, False] == m.kept
    m.trim_incomplete(wanted=["a", "b", "c"], strict=True)
    assert [False, False, False] == m.kept


def test_trim_to_range_min(full):
    full.trim_to_range("zpe", minimum=-235)
    assert [True, True, True, True, False, False, False] == full.kept


def test_trim_to_range_previously_trimmed(full):
    full.kept = [False, True, True, True, True, True, True]
    full.trim_to_range("zpe", minimum=-235)
    assert [False, True, True, True, False, False, False] == full.kept


def test_trim_to_range_max(full):
    full.trim_to_range("zpe", maximum=-225)
    assert [False, False, False, True, True, True, True] == full.kept


def test_trim_to_range_min_max(full):
    full.trim_to_range("zpe", minimum=-255, maximum=-205)
    assert [False, True, True, True, True, True, False] == full.kept


def test_trim_to_range_errors(full):
    with pytest.raises(TypeError):
        full.trim_to_range("command")
    with pytest.raises(AttributeError):
        full.trim_to_range("zpe", attribute="bla")
    with pytest.raises(ValueError):
        full.trim_to_range("freq")


@pytest.mark.parametrize("trimmed", [True, False])  # if really empty or fully trimmed
@pytest.mark.parametrize(
    "method,args",
    [
        ("trim_not_optimized", []),
        ("trim_imaginary_frequencies", []),
        ("trim_non_matching_stoichiometry", []),
        ("trim_non_normal_termination", []),
        ("trim_inconsistent_sizes", []),
        ("trim_incomplete", []),
        ("trim_to_range", ["gib"]),
        ("trim_rmsd", [1, 1]),
    ],
)
def test_trim_empty(empty, single, method, args, trimmed):
    confs = empty if not trimmed else single
    if trimmed:
        confs.kept[0] = False
    callable = getattr(confs, method)
    callable(*args)
    assert confs.kept == ([False] if trimmed else [])


def test_select_all(full):
    full.kept = [True, False, True, False, True, False, False]
    assert not all(full.kept)
    full.select_all()
    assert all(full.kept)


def test_reject_all(full):
    full.kept = [True, False, True, False, True, False, False]
    full.reject_all()
    assert not all(full.kept)


def test_kept_raises(empty, full):
    with pytest.raises(TypeError):
        full.kept = 1
    with pytest.raises(TypeError):
        full.kept = {"bla": 1}
    with pytest.raises(KeyError):
        full.kept = ["zero"]
    with pytest.raises(ValueError):
        full.kept = [True] * 2
    with pytest.raises(ValueError):
        empty.kept = [True] * 2
    with pytest.raises(ValueError):
        full.kept = [True] * 20
    with pytest.raises(IndexError):
        full.kept = [200]
    with pytest.raises(TypeError):
        full.kept = [[]]


def test_kept(empty, full):
    assert [True] * 7 == full.kept
    full.kept = "imag stoich term".split(" ")
    assert [False, False, True, True, True, False, False] == full.kept
    full.kept = [1, 3, 5]
    assert [False, True, False, True, False, True, False] == full.kept
    full.kept = [True, False, False, True, False, False, True]
    assert [True, False, False, True, False, False, True] == full.kept
    full.kept = False
    assert [False] * 7 == full.kept
    full.kept = True
    assert [True] * 7 == full.kept
    assert [] == empty.kept
    empty.kept = []
    assert [] == empty.kept
    empty.kept = False
    assert [] == empty.kept


def test_by_index(full):
    for i, n in enumerate(full.filenames):
        assert full.by_index(i) is full[n]


def test_index_key(full):
    for i in range(len(full)):
        assert full.index_of(full.key_of(i)) == i


@pytest.mark.parametrize(
    "genre,output",
    [("freq", True), ("scf", True), ("excess", True), ("nohave", False)],
)
def test_has_genre(jagged, genre, output):
    assert jagged.has_genre(genre) == output


@pytest.mark.parametrize(
    "genre,output",
    [("freq", True), ("scf", True), ("excess", False), ("nohave", False)],
)
def test_has_genre_trimmed(jagged, genre, output):
    jagged.kept = list(range(4))
    assert jagged.has_genre(genre, ignore_trimming=False) == output


@pytest.mark.parametrize(
    "genre,output",
    [("freq", True), ("scf", True), ("excess", True), ("nohave", False)],
)
def test_has_genre_untrimmed(jagged, genre, output):
    jagged.kept = list(range(4))
    assert jagged.has_genre(genre, ignore_trimming=True) == output


@pytest.mark.parametrize(
    "genre,output",
    [
        (["freq", "nohave"], True),
        (["scf", "nohave"], True),
        (["excess", "nohave"], True),
        (["nohave", "alsonohave"], False),
    ],
)
def test_has_any_genre(jagged, genre, output):
    assert jagged.has_any_genre(genre) == output


@pytest.mark.parametrize(
    "genre,output",
    [
        (["freq", "nohave"], True),
        (["scf", "nohave"], True),
        (["excess", "nohave"], False),
        (["nohave", "alsonohave"], False),
    ],
)
def test_has_any_genre_trimmed(jagged, genre, output):
    jagged.kept = list(range(4))
    assert jagged.has_any_genre(genre, ignore_trimming=False) == output


@pytest.mark.parametrize(
    "genre,output",
    [
        (["freq", "nohave"], True),
        (["scf", "nohave"], True),
        (["excess", "nohave"], True),
        (["nohave", "alsonohave"], False),
    ],
)
def test_has_any_genre_untrimmed(jagged, genre, output):
    jagged.kept = list(range(4))
    assert jagged.has_any_genre(genre, ignore_trimming=True) == output


@pytest.mark.parametrize(
    "genre,output",
    [
        (["freq", "iri"], True),
        (["scf", "iri"], False),
        (["excess", "iri"], False),
        (["nohave", "iri"], False),
    ],
)
def test_all_have_genres(jagged, genre, output):
    assert jagged.all_have_genres(genre) == output


@pytest.mark.parametrize(
    "genre,output",
    [
        (["freq", "iri"], True),
        (["scf", "iri"], True),
        (["excess", "iri"], False),
        (["nohave", "iri"], False),
    ],
)
def test_all_have_genres_trimmed(jagged, genre, output):
    jagged.kept = list(range(4))
    assert jagged.all_have_genres(genre, ignore_trimming=False) == output


@pytest.mark.parametrize(
    "genre,output",
    [
        (["freq", "iri"], True),
        (["scf", "iri"], False),
        (["excess", "iri"], False),
        (["nohave", "iri"], False),
    ],
)
def test_all_have_genres_untrimmed(jagged, genre, output):
    jagged.kept = list(range(4))
    assert jagged.all_have_genres(genre, ignore_trimming=True) == output


def test_inconsistency_allowed_context(single):
    with single.inconsistency_allowed:
        assert single.allow_data_inconsistency
    assert not single.allow_data_inconsistency


def test_trimmed_to_context(full):
    with full.trimmed_to("imag stoich term".split(" ")):
        assert [False, False, True, True, True, False, False] == full.kept
    assert [True] * 7 == full.kept


def test_untrimmed_context(full):
    full.kept = [False, False, True, True, True, False, False]
    with full.untrimmed:
        assert [True] * 7 == full.kept
    assert full.kept == [False, False, True, True, True, False, False]


# TODO: add tests for View objects
